/*! @file
 * @brief
 * implementation of (invariant) CRelation methods
 * that can be called by servers and clients
 */

#include <string>
#include <iostream>
#include <cmath>
#include <gmp.h>
#include "utils.H"

#include "qsieve.H"

using namespace std;

extern int PrimeNumbers[StaticFactorbase::MaxSize];
extern mpz_t n;


ostream& operator<< (ostream& ostr, const CRelation& GL)
{
  if (GL.Relation_dense) ostr << "dense"; else ostr << "sparse";
  ostr << "1=";
  
  char *str;
  str = new char [mpz_sizeinbase(GL.Delta,10)+2]; mpz_get_str(str, 10, GL.Delta);
  ostr << str << "^2";
  delete [] str; 
  
  if (GL.Relation_dense)
    { 
      for (int i=GL.Relation_dense->first(); i>=0; i=GL.Relation_dense->next(i))
	ostr << "*p" << i;
    }
  else
    {
      const CTinyFBsizetypeVector& R = *GL.Relation_sparse;
      for (int pos=0; pos<R.size(); ++pos)
	ostr << "*p" << R[pos];
    }
  
  return ostr;
}


void CRelation::convert_Relation_to_dense()
{
  //cout << "Converting relation from sparse to dense representation." << endl;
  if (Relation_dense)
    { cerr << "Relation is already dense!" << endl; return; }
  Relation_dense = new myBitString;
  for (int pos=Relation_sparse->size()-1; pos>=0; --pos)
    Relation_dense->set((*Relation_sparse)[pos]);
  //Relation_sparse->erase(Relation_sparse->begin(),Relation_sparse->end());
  delete Relation_sparse; Relation_sparse=NULL; // Sparsedarstellung lschen
}

void CRelation::convert_Relation_to_sparse()
{
  //cout << "Converting relation from dense to sparse representation." << endl;
  if (Relation_sparse)
    { cerr << "Relation is already sparse!" << endl; return; }
  Relation_sparse = new CTinyFBsizetypeVector(Relation_dense->count(1));
  for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
    Relation_sparse->fast_append(i);
  //Relation_dense->clear();
  delete Relation_dense; Relation_dense=NULL; // Densedarstellung lschen
}


void CRelation::combine(const CRelation& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Kombiniere Gleichungszeilen..." << endl;
  //cout << "Gl.1: "; SanityCheck();
  //cout << "Gl.2: "; GL2.SanityCheck();
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen
      
      for (int i=GL2.Relation_dense->last(); i>=0; i=GL2.Relation_dense->prev(i))
	if (Relation_dense->test(i))
	  if (i>0)
	    {
	      mpz_mul_ui(x,x,PrimeNumbers[i]);
	      if (mpz_cmp(x,n)>=0)
		{
		  mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		  mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		}
	    } //else mpz_neg(Delta,Delta);
      
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	for (int pos=0; pos<GL2.Relation_sparse->size(); ++pos)
	  {
	    if (Relation_dense->test((*GL2.Relation_sparse)[pos]))
	      if ((*GL2.Relation_sparse)[pos]>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[(*GL2.Relation_sparse)[pos]]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	    Relation_dense->invert((*GL2.Relation_sparse)[pos]);
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
	neueMenge.reset();
        const CTinyFBsizetypeVector& R = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto fertig;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto fertig;
	      }
	    while (R2[p2]==R[p1])
	      {
		if (R[p1]>0)
		  {
		    mpz_mul_ui(x,x,PrimeNumbers[R[p1]]);
		    if (mpz_cmp(x,n)>=0)
		      {
			mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
			mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		      }
		  } //else mpz_neg(Delta,Delta);
		++p1; ++p2;
		if (p2==R2.size()) goto fertig;
		if (p1==R.size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
        //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
      }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Relation; dieses also dazumultiplizieren!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  relevant_factor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}

// Routinen, die ein mehrfaches Kombinieren beschleunigen:
// NEU: Multikombinationsroutinen sollten das Kombinieren einer
//      Relation mit vielen anderen Relationen beschleunigen, da
//      das Update der Faktorbasis-Exponenten der Relation
//      an das Ende aller Verknpfungen geschoben werden kann und
//      nicht mehr simultan durchgefhrt werden mu.

CRelation::TExponentArrayElement CRelation::ExponentArray[StaticFactorbase::MaxSize+32] = {0};
 // +32 because we want to allow fast routines to avoid costly range checks...

unsigned int CRelation::multi_combine_Counter = 0;
 
void CRelation::multi_combine_init()
{
  // Initialisierung des Exponentenvektors
  // Hinweis: Der Aufruf dieser Routine darf auch entfallen,
  //          sofern sichergestellt ist, dass der letzte
  //          multi_combine_...-Aufruf ein  multi_combine_exit() war.
  if (!multi_combine_Counter) return;
  multi_combine_Counter=0;
  for (int i=0; i<StaticFactorbase::Size(); i++) ExponentArray[i]=0;
}

void CRelation::multi_combine_exit()
{
  if (!multi_combine_Counter) return; // gibt nichts zu tun...
#ifdef VERBOSE
  cout << "Multicombine_exit after " << multi_combine_Counter
       << " calls." << endl;
#endif

  if (multi_combine_Counter>std::numeric_limits<TExponentArrayElement>::max())
   {
     MARK;
     cerr << "Elements of CRelation::ExponentArray may have exceeded limits," << endl
          << "you have to increase the type of CRelation::TExponentArrayElement" << endl
          << "and recompile the program to assure correct computation." << endl
          << "(Data computed so far are still correct and can be reused afterwards.)" << endl;
     exit(8);
   }

  optisize(); // optimize Size of vector (sparse<->dense, etc.)

  // nun die Faktor-Exponenten in das Delta ziehen:
  mpz_set_ui(y,1); // Produkt hierin sammeln
  ExponentArray[0]=0; // Primzahl "-1" immer unbercksichtigt lassen

  /* 
     Idee zur Beschleunigung:
     Exponentenfeld absteigend nach Exponenten sortieren, anschlieend
     das Produkt analog zum "Horner-Schema" auswerten.
     Damit das Sortieren nicht zu teuer wird, machen wir eine Art Bucketsort,
     wobei das "Flag[i]" angibt, ob es zu i einen Exponenten gibt.
     Alle zu einem Exponenten e gehrenden Basen werden dann in "Basis[e]"
     multiplikativ zusammengefat.
  */

  bool* const Flag = new bool[StaticFactorbase::Size()];
  for (int i=0; i<StaticFactorbase::Size(); ++i) Flag[i]=false; // initialisieren
  mpz_t* const Basis = new mpz_t[StaticFactorbase::Size()];

  for (unsigned int i=1; i<StaticFactorbase::Size(); ++i)
   if (ExponentArray[i])
    {
      const unsigned int e=ExponentArray[i];
      ExponentArray[i]=0; // Exponenten wieder zurcksetzen

      if (e>=StaticFactorbase::Size())
       {
         cout << "Unexpected exponent " << e << "! Use normal exponentiation..." << endl;
         mpz_set_ui(y,PrimeNumbers[i]); mpz_powm_ui(y,y,e,n);
         mpz_mul(Delta,Delta,y);
         continue;
       }

      if (Flag[e])
       {
         // Basis erweitern
         mpz_mul_ui(Basis[e],Basis[e],PrimeNumbers[i]);
         if (mpz_cmp(Basis[e],n)>=0)
          { 
            //cout << "Basis erweitern, mod sinnvoll..." << endl;
            mpz_mod(Basis[e],Basis[e],n);
          }
       }
      else 
       { 
         // Basis initialisieren
         Flag[e]=true;
         mpz_init(Basis[e]); mpz_set_ui(Basis[e],PrimeNumbers[i]);
       }
    }

 /*
  Produkt der Basen analog zum Horner-Schema auswerten,
  Beispiel: a^7 * b^4 * c^3 * d 
            = a^(7-4) * (a*b)^(4-3) * (a*b*c)^(3-1) * (a*b*c*d)
 */     

  unsigned int e2 = StaticFactorbase::Size()-1;
  Flag[0]=true; // sicherheitshalber fr nachfolgende Schleife...
  while (!Flag[e2]) --e2; // hchsten gesetzten Exponenten holen

  mpz_set_ui(y,1);
  if (e2>1)
   for (unsigned int e=e2-1; e>0; --e)
    if (Flag[e])
     {
       mpz_mul(Basis[e],Basis[e],Basis[e2]);
       if (mpz_cmp(Basis[e],n)>=0)
        {
          //cout << "Auswertung A, mod sinnvoll" << endl; 
          mpz_mod(Basis[e],Basis[e],n);
        }
       if (e2-e>1) mpz_powm_ui(Basis[e2],Basis[e2],e2-e,n);
       mpz_mul(y,y,Basis[e2]);
       if (mpz_cmp(y,n)>=0)
        {
          //cout << "Auswertung B, mod sinnvoll" << endl; 
          mpz_mod(y,y,n);
        }
       mpz_clear(Basis[e2]); // do not forget: mpz-Zahl wieder freigeben
       e2=e;
     }

  if (e2>1) mpz_powm_ui(Basis[e2],Basis[e2],e2,n);
  if (e2>0)
   {
      mpz_mul(y,y,Basis[e2]); mpz_mod(y,y,n);
      mpz_clear(Basis[e2]); // do not forget: mpz-Zahl wieder freigeben
   }

  delete [] Basis; delete [] Flag;
  mpz_mul(Delta,Delta,y); mpz_mod(Delta,Delta,n);
  multi_combine_Counter=0;
#ifdef VERBOSE
  cout << "Multi_combine_exit finished." << endl;
#endif
}

void CRelation::multi_combine_main(const CRelation& GL2)
  // bergabe von GL2 ber Referenz, um effizienter zu sein...
  /* symmetrische Differenzen ermitteln */
{
  //cout << "Multi-Kombiniere Relationen..." << endl;
  ++multi_combine_Counter; // Anzahl Aufrufe zhlen

  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Exponentenfeld vermerken
  */

  if ( GL2.Relation_dense ) /* GL2.Relation is dense */
    {
      if (!Relation_dense) // Relation ist nicht dense
	convert_Relation_to_dense(); // Relation in dichte Darstellung bringen

      /* Zunchsten werden die Bits der beiden Relationen
         (temporr!) komponentenweise durch "and" verknpft.
         Dadurch werden die Primzahlen ermittelt, die verknpft Quadrate
         ergeben. Diese knnen dann im Exponentenfeld bercksichtigt werden.
         In gewisser Weise wird die Bitstrings komponentenweise wie
         mit einem "Halbaddierer" verknpft.
          Additionsergebnis (durch xor) berschreibt anschlieend den
          Bitstring, whrend der bertrag im Expoentenfeld addiert wird.
       */
      Relation_dense->test_and_add_carry(*GL2.Relation_dense,ExponentArray);
      (*Relation_dense)._xor(*GL2.Relation_dense); // symmetrische Differenz bilden
    }
  else /* GL2.Relation is sparse */
    if (Relation_dense) /* and Relation is dense */
      { /* symmetrische Differenz in dense-Darstellung bilden,
	   wobei der andere Vektor sparse ist */
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	for (int pos=0; pos<R2.size(); ++pos)
	  {
	    if (Relation_dense->test_and_invert(R2[pos])) ++ExponentArray[R2[pos]];
	  }
      }
    else
      { /* symmetrische Differenz in sparse-Darstellung bilden */
	int p1 = 0;
	int p2 = 0;
	static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
	neueMenge.reset();
	const CTinyFBsizetypeVector& R  = *Relation_sparse;
	const CTinyFBsizetypeVector& R2 = *GL2.Relation_sparse;
	while (p2<R2.size() && p1<R.size())
	  {
	    while (R2[p2]<R[p1])
	      {
		neueMenge.fast_append(R2[p2]); ++p2;
		if (p2==R2.size()) goto fertig;
	      }
	    while (R[p1]<R2[p2])
	      {
		neueMenge.fast_append(R[p1]); ++p1;
		if (p1==R.size()) goto fertig;
	      }
	    while (R2[p2]==R[p1])
	      {
		++ExponentArray[R[p1]];
		++p1; ++p2;
		if (p2==R2.size()) goto fertig;
		if (p1==R.size()) goto fertig;      
	      }
	  }
      fertig:
	while (p2!=R2.size()) { neueMenge.fast_append(R2[p2]); ++p2; }
	while (p1!=R.size()) { neueMenge.fast_append(R[p1]); ++p1; }
	Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
      }
  
  // nicht vergessen: Das neue Delta beinhaltet natrlich auch das
  // alte Delta der anderen Relation; dieses also dazumultiplizieren!
  mpz_mul(Delta,Delta,GL2.Delta); mpz_mod(Delta,Delta,n);
  
  //-------------------------------------------------------------------
  // WICHTIG: Nach dem letzten Aufruf von multi_combine_main
  // mu die Relation mit dem Exponentenfeld abgeglichen werden!!!
  // Hierzu ist ein abschlieendes multi_combine_exit aufzurufen!
  //-------------------------------------------------------------------

  relevant_factor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
}



CmpqsFactor CRelation::combine(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  CmpqsFactor f=combine(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}

CmpqsFactor CRelation::multi_combine_main(istream &in, const streampos pos)
{
  const streampos merkfpos=in.tellg(); // ursprngliche Fileposition merken
  in.seekg(pos);
  CmpqsFactor f=multi_combine_main(in);
  in.seekg(merkfpos); // ursprngliche Fileposition wiederherstellen
  return f;
}


int GetValue(istream &in)
{
  static int ahead_value = 0;
  if (ahead_value>0) { int w=ahead_value; ahead_value=0; return w; }

  char c;
  in >> c;
  if ( c>='A' && c<='Z' ) return c-('A'-1); // A=1,...,Z=26
  if ( c>='g' && c<='z' )
   {
     int h = (c-'g');
     ahead_value=(h%5)+1; // nchster "vorausgelesener" Wert
     return (h/5)+1; 
   }
  if (c=='0') return 0; // da keine fhrenden Nullen
  if (c=='$') return -1;
  if (c=='%') return -2;

  if (c==' ')
   {
     MARK;
     cerr << "GetValue: Read a space? How to handle this?" << endl;
     exit(1);
   }

  // ansonsten normal gespeicherte Zahl:
  int w;
  in.unget();
  in >> w;
  return w;
}

void PutValue(ostream &out, const int w)
{
  static int prev_value = 0;
  if (w<=5 && w>=1 && prev_value>0) // dann kann kodiert werden
    {
      out << char( (prev_value-1)*5 + w-1 + 'g' );
      prev_value=0;
      return;
    }

  if (prev_value>0)
   {
     out << char(prev_value+('A'-1)); // Wert ausgeben
     prev_value=0;
   }

  if (w>26) 
   {
     out << w << " "; // Wert unkodiert ausgeben
     return;
   }
  if (w>0)
   {
     if (w<=4) { prev_value=w; return; } // nchstes Mal abwarten
     out << char(w+('A'-1)); // Wert als char kodiert ausgeben
     return;
   }
  switch (w)
   {
     case -2: out << "%"; break;
     case -1: out << "$"; break;
     case  0: out << "0"; break;
     default:
      MARK;
      cerr << "PutValue: don't know how to handle this..." << endl;
      cerr << "w=" << w << endl;
      exit(1);
   }
}

void CRelation::SanityCheck() const
{
  mpz_set_ui(y,1); 
  if (Relation_sparse)
    {
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
	  if ((*Relation_sparse)[i]>0) 
	    mpz_mul_ui(y,y,PrimeNumbers[(*Relation_sparse)[i]]);
	  else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  else
    {
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
	  if (i>0) mpz_mul_ui(y,y,PrimeNumbers[i]); else mpz_neg(y,y);
	  mpz_mod(y,y,n);
	}
    }
  
  mpz_invert(x,Delta,n); mpz_powm_ui(x,x,2,n);
  
  cout << "checking for congruency: ";
  cout << x << " = " << y << " (mod N)" << endl;
  if (mpz_cmp(x,y)!=0) 
    {
      cout << "Error! Values are NOT congruent!" << endl;
      exit(1);
    }
  
}
