/*! @file
 * @brief
 * mpqs sieving stuff, header file
 */

#ifndef SIEVING_HEADER_
#define SIEVING_HEADER_

#include <qsieve-fwd.H>
#include <cmath>


/*!
 * @short
 * size of sieve interval in cacheable memory for fast access
 *
 * The logical sieve interval is sieved in separate smaller parts, because
 * memory access in cache can be up to 200 times faster than uncached memory
 * access (cost of a cachemiss). The value of @p PhysicalSieveSize is
 * ideally the amount of memory that fits into processor cache while not
 * replacing too much other cacheable data.
 */
const int PhysicalSieveSize = 64*1024; // =65536
/* In (physikalischen) Teilintervallen von "PhysicalSieveSize" werden die Relationen
   im (logischen) Siebintervall [-M,M] (siehe auch Deklaration von M) gesiebt.
   Es handelt sich deshalb um eine bereits zur Compilierungszeit feststehende Konstante.

   Hinweis: Je grer "PhysicalSieveSize", desto effizienter geht das Aussieben
   der Relationen. Gleichzeitig erhht sich aber auch der Speicherbedarf. 
   Der Einflu der Siebgre ist allerdings geringer als der der gewhlten
   Primzahlbasis. Ein Wert um 1 Million drfte daher vollkommen ausreichen. 
   Andererseits kann mglicherweise im Cachespeicher des Rechners gesiebt
   werden, wenn die Siebgre klein gewhlt wird. Kleine Siebgren erhhen
   zwar den rechnerischen Overhead, aber dafr ist der Zugriff auf das Sieb
   mglicherweise deutlich schneller. Cachemisses knnen je nach Level
   einzelne Zugriffe um den Faktor 10 bis 200 verteuern! Deshalb sollte man
   auch eine Siebgre um 100000 (bei 512K Cache) ausprobieren.

   Fr eventuellen optimierten Zugriff ber Assemblerbefehle sollte
   PhysicalSieveSize*TSIEBELEMENTSIZE ein Vielfaches der Maschinenwortgre (bei
   32-Bit-Rechnern also 4, bei 64-Bit-Rechnern 8; beim AMD-Athlon wegen der
   MMX-Untersttzung und wegen der Cacheline: 64) sein.  Idealerweise whle
   man ein Vielfaches von 1024, dann ist man auf der sicheren Seite.  (Das
   stellt keine wirkliche Einschrnkung dar, ermglicht es aber, das Sieb
   effizient zu initialisieren ohne Sonderflle beachten zu mssen. -- Wenn
   Assembler-Code aktiviert wird, verlasse ich mich auf diesen Constraint!)
*/


extern int LogicalSieveSize; // das zu jedem Polynom gehrige Siebintervall ist [-LogicalSieveSize,LogicalSieveSize]
/*
   Constraint for AMD-3DNow! specific code: abs(LogicalSieveSize)<=2^23 (because of float precision)

   LogicalSieveSize ist ein im Unterschied zu "PhysicalSieveSize" ein
   logischer Wert, der weniger stark von der verwendeten Hardware abhngig
   ist. Die logische Intervallgre ist vor allem abhngig von der Gre der
   zu faktorisierenden Zahl zu whlen.  Der hier festgelegte Wert ist also
   lediglich ein Defaultwert, der von der Konfiguration vor Beginn eines
   konkreten Faktorisierungslaufs (zur Laufzeit) modifiziert werden kann
   (und soll).
   Bemerkung: In multiple-polynomial-quadratic-sieve-Varianten sollte man
   mglichst kleine Siebintervalle whlen, da die relative Trefferquote von
   Relationen mit der Intervallgre abnimmt. (In 10 kleinen Intervallen ist
   die Chance, eine Relation zu finden, hher als in einem 10mal so groen
   Intervall).
   Gegen kleine Intervalle spricht jedoch, dass Kosten fr das Wechseln der
   Polynome, Neuinitialisierungen und Wurzelberechnungen anfallen.
*/


// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
// !!! WICHTIG !!! #define und typedef mssen einander entsprechen !!!
// !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#define TSIEBELEMENTSIZE 1 /* 1 = signed char, 2 = short signed int, 4 = signed int, 0 = no assembler inline code (=generic code) */
// obiges define sorgt bei Verwendung der ASM386-Direktive fr inline-assembler-code in der Siebroutine

#if TSIEBELEMENTSIZE == 0
 typedef signed char TSieveElement; // must be signed!
 #define log_Multiplikator 1 /* Siebwert wird nach damit nach TSiebelement konvertiert */
 #ifdef ASM386
  #warning "assembly code is disabled for sieving!"
 #endif
#elif TSIEBELEMENTSIZE == 1
 typedef signed char TSieveElement; // must be signed!
 #define log_Multiplikator 1 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#elif TSIEBELEMENTSIZE == 2
 typedef signed short int TSieveElement; // must be signed!
 #define log_Multiplikator 16 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#elif TSIEBELEMENTSIZE == 4
 typedef signed int TSieveElement; // must be signed!
 #define log_Multiplikator 256 /* Siebwert wird nach damit nach TSiebelement konvertiert */
#else
 #error "invalid value for TSIEBELEMENTSIZE"
#endif

/* Je kleiner der Typ von TSieveElement ist, desto 
    a) kleiner ist der fr das Sieb bentigte Speicher,
    b) schneller kann deshalb gesiebt werden.
   Allerdings ist die Granularitt zum Erkennen von Relationen bei
   kleineren Datentypen grber und deshalb schlechter.
   Auerdem sind prozessorspezifische Effekte (z.B. Alignment-Penalties)
   zu bercksichtigen!
   In Frage kommen:
    - signed char (1 Byte)
    - short signed int (2 Byte)
    - signed int (4 Byte)
    - float, double (Fliekommatypen, mehrere (>3) Bytes)
   Die Bytegren knnen zwischen Systemen variieren. Ich habe aber nicht
   geprft, ob das Programm berhaupt auf mit einer Systemeinstellung von
   int < 4 Byte, short int < 2 Byte lauffhig ist.

   Wichtiger Hinweis: Wenn man Fliekommatypen als TSieveElement verwenden will (was ich nicht mehr empfehle),
   dann sind im Programm die "ceil"-Aufrufe fr TSieveElement kontraproduktiv und sollten entfernt werden.
   Aber es gibt keinen vernnftigen Grund mehr, mit Fliekommatypen zu sieben.
*/

// #define log_Multiplikator <VALUE> /* Siebwert wird nach damit nach TSiebelement konvertiert */
/*
  -> refer TSIEBELEMENTSIZE
  Da die Vorkommastellen der natrliche Logarithmen von Primzahlen
  zu hohe Ungenauigkeiten beim Sieben verursachen, werden sie noch
  mit log_Multiplikator multipliziert, um eine hhere Genauigkeit
  zu erreichen.
  Der Multiplikator kann auf 1 gesetzt sein, wenn TSieveElement Fliekommatyp
  ist. Ansonsten sollte man ihn mglichst gro whlen, und zwar so, dass
  der Threshold der blicherweise zu faktorisierenden Zahlen noch in den
  Datentyp TSieveElement pat.
  Damit durch Abrundungen keine Relationen durch die Lappen gehen, werden
  aber beim Sieben grundstzlich aufgerundete Werte abgezogen, so dass ein
  kleiner Multiplikator hier allenfalls ein zu engmaschiges Sieb erzeugt.
  Die exakten Auswirkungen verschiedener Einstellungen wren allerdings
  noch genauer zu untersuchen.
  Wenn man nicht genau wei, was man tut, sollte man eher TSieveElement
  "hhertypen". Der Multiplikator sollte dann deutlich ber 1 liegen (10 - 1000).
*/



#ifdef IS_SERVER

 // The server does no sieving,
 // but the server source code needs some information about how sieving
 // is done. Additionally the server sources contain some code that is
 // depending on data structures defined here. We could eliminate these
 // lines by conditional compiling (#ifdef IS_SERVER), but this would
 // be to bloated. Instead we provide basic dummy classes which produce
 // dead code, if compiled in SERVER-mode.

template <typename T> class CDummyArray
{
 private:
  T dummy;
 public:
  inline T& operator[] (const int)
   {
     return dummy;
   }
  inline const T& operator[] (const int) const
   {
     return dummy;
   }
};

class SieveControl
{
 private:
  static const int PrimzahlStartIndex = 5;
  /* ab diesem Primzahlindex wird gesiebt.
     Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
     der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
     beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
     so einfach) gesiebt werden.
     Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
     grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
     viele Relationen durchs Sieb fallen...
   */

  inline static const double DirtyFactor(const int i)
   {
     if (i<PrimzahlStartIndex+4) return 1.0;
     return 1.0/sqrt(i); // irgendeine langsam gegen 0 strebende Funktion
   }
  /* Der DirtyFactor gibt die Wichtung an, mit der die i-te Primzahl der Faktorbasis,
     die "dirty" sind, als Siebtreffer auf das Sieb aufgeschlagen werden.
     Beispiel:
      Wenn p normal siebt, gibt es im Sieb an jeder p-ten Stelle einen Treffer, der
      voll gewichtet wird, z.B. mit log(p).
      Wenn p "dirty" ist, gibt es an *allen* Siebstellen einen Treffer fr p, der
      mit DirtyFactor skaliert ist, also z.B. DirtyFactor*log(p). Erst wenn
      spter eine potentielle Relation erkannt wird, dann wird der hierdurch
      verursachte Fehler korrigiert.
     Ein DirtyFactor von 1.0 verhlt sich gegenber einem normalen Siebvorgang
     bis auf Rundungseffekte neutral, beschleunigt aber den Siebvorgang.
     Je kleiner der DirtyFactor ist, desto mehr Primzahlen knnen im Laufe des
     Regelungsvorgangs "dirty" werden. (Die Erkennungsrate pro Siebintervall
     sinkt dann allerdings zugunsten der Siebgeschwindigkeit.)
     Der DirtyFactor 1/n verhlt sich grob gesprochen etwa so, als wrde jede
     n-te der "dirty" Primzahlen "automatisch" in jeder Relation vorhanden sein.
   */

  static const int sieben_ab_Primzahlindex(void) { return PrimzahlStartIndex; }

  static CDummyArray<TSieveElement> log_PrimeNumbers;
  static CDummyArray<TSieveElement> log_PrimeNumbers_dirty;

 public:
  friend void compute_StaticFactorbase();
};

CDummyArray<TSieveElement> SieveControl::log_PrimeNumbers;
CDummyArray<TSieveElement> SieveControl::log_PrimeNumbers_dirty;

CDummyArray<TSieveElement> log_Primzahl_of_PrimePowers;


#else /* IS_SERVER not defined */

// these are the "full" declarations
#include <iosfwd>
#include <gmp.h>
#include "utils.H"
#include "mpz_wrapper.H"
using namespace my_mpz_wrapper;

using std::ostream;
using std::istream;

using std::cout;
using std::cerr;


extern int SieveOffset;
/* Im allgemeinen wird das fr ein Polynom zu siebende Siebintervall 
   grer sein als die (physikalisch optimale) Siebgre.
   Das Siebintervall ist daher in Teilstcken zu sieben. Hierfr wird der
   Sieboffset bentigt, der den Startpunkt des jeweiligen Teilsiebs angibt.
 */


extern TSieveElement SieveArray_[PhysicalSieveSize+64+8] __attribute__ ((aligned (64))); // Speicher fr das Sieb
TSieveElement* const SieveArray = &SieveArray_[64];
// SieveArray[-64..-1] are -1 (for fast value-based end-of-loop range detections)
// SieveArray[0..PhysicalSieveSize-1] contain actually sieve values
// SieveArray[PhysicalSieveSize..PhysicalSieveSize+8] contain dummy values


extern TSieveElement log_Primzahl_of_PrimePowers[StaticFactorbase::max_additional_Powers];
/* fr weitere Primzahlpotenzen stehen hier die konvertierten Logarithmen der Primzahlen */


extern int PrimeNumbers[StaticFactorbase::MaxSize];
extern int Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];
extern mpz_t n, kN, r;
extern double Factor_Threshold;
extern int biggest_Prime_in_Factorbase;
extern int NumberOf_more_PrimePowers;
extern int PrimePowers[StaticFactorbase::max_additional_Powers];
extern int Delta_of_PrimePowers[StaticFactorbase::max_additional_Powers][2];


inline TSieveElement log_DynamicLargePrime (const int p)
{
#if defined(ASM386) && TSIEBELEMENTSIZE == 1 && log_Multiplikator == 1
 #ifdef DEBUG
  #warning "assembler i386 TSIEBELEMENT log lookup enabled"
 #endif
  // we assume 8103<p<=1318815734 (=e^9)
  register TSieveElement logval;
  asm ( \
       "movb $21,%[L] \n\t" \
       "cmpl $485165195,%[P] # exp(20) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $178482300,%[P] # exp(19) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $65659969,%[P] # exp(18) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $24154952,%[P] # exp(17) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $8886110,%[P] # exp(16) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $3269017,%[P] # exp(15) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $1202604,%[P] # exp(14) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $442413,%[P] # exp(13) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $162754,%[P] # exp(12) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $59874,%[P] # exp(11) \n\t" \
       "sbbb $0,%[L] \n\t" \
       "cmpl $22026,%[P] # exp(10) \n\t" \
       "sbbb $0,%[L]" \
       : [L] "=&q" (logval) : [P] "r" (p) : "cc");
#ifdef DEBUG
  // remark: do not use std::log(), as it might be ambiguous for some
  // versions of g++ (-> see utils.H)
  if (logval!=ceil(log_Multiplikator*log(p)))
   {
     MARK;
     cout << "P: " << p << " log:" << static_cast<int>(logval) << "  " << ceil(log_Multiplikator*log(p)) << endl;
     exit(1);
   }
#endif
  return logval;
#else
  // generic version
  return static_cast<TSieveElement>(ceil(log_Multiplikator*log(p)));
#endif
}


/// provide adaptive control mechanism to improve sieving
class SieveControl
{
 private:
  static const int PrimzahlStartIndex = 5;
  /* ab diesem Primzahlindex wird gesiebt.
     Der Wert mu grer als 1 sein, da die ersten beiden Faktoren
     der Primzahlbasis -1 und mglicherweise die 2 sind. Fr diese
     beiden Zahlen sowie dem Vorfaktor von kN kann nicht (jedenfalls nicht
     so einfach) gesiebt werden.
     Fr ein schnelles, aber unsauberes dirty-sieving kann auch ein
     grerer Wert gewhlt werden. Aber zu hohe Werte lassen dann auch
     viele Relationen durchs Sieb fallen...
   */

  inline static const double DirtyFactor(const int i)
   {
     if (i<PrimzahlStartIndex+4) return 1.0;
     return 1.0/sqrt(i); // irgendeine langsam gegen 0 strebende Funktion
   }
  /* Der DirtyFactor gibt die Wichtung an, mit der die i-te Primzahl der Faktorbasis,
     die "dirty" sind, als Siebtreffer auf das Sieb aufgeschlagen werden.
     Beispiel:
      Wenn p normal siebt, gibt es im Sieb an jeder p-ten Stelle einen Treffer, der
      voll gewichtet wird, z.B. mit log(p).
      Wenn p "dirty" ist, gibt es an *allen* Siebstellen einen Treffer fr p, der
      mit DirtyFactor skaliert ist, also z.B. DirtyFactor*log(p). Erst wenn
      spter eine potentielle Relation erkannt wird, dann wird der hierdurch
      verursachte Fehler korrigiert.
     Ein DirtyFactor von 1.0 verhlt sich gegenber einem normalen Siebvorgang
     bis auf Rundungseffekte neutral, beschleunigt aber den Siebvorgang.
     Je kleiner der DirtyFactor ist, desto mehr Primzahlen knnen im Laufe des
     Regelungsvorgangs "dirty" werden. (Die Erkennungsrate pro Siebintervall
     sinkt dann allerdings zugunsten der Siebgeschwindigkeit.)
     Der DirtyFactor 1/n verhlt sich grob gesprochen etwa so, als wrde jede
     n-te der "dirty" Primzahlen "automatisch" in jeder Relation vorhanden sein.
   */  

  static TSieveElement log_PrimeNumbers[StaticFactorbase::MaxSize];
  static TSieveElement log_PrimeNumbers_dirty[StaticFactorbase::MaxSize];
    // Hier stehen die konvertierten Logarithmen der Primzahlen aus der Faktorbasis

  static TSieveElement log_Threshold; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert
  static int _sieben_ab_Primzahlindex;

  static signed int DirtynessPegel;

public:
  inline static const TSieveElement PrimzahlSiebWichtung(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck; blicherweise ist
     // dies der (skalierte) Logarithmus, der in der Tabelle abgelegt wurde. 
     // Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers[i];
   }
  inline static const TSieveElement PrimzahlSiebWichtung_dirty(const int i)
   {
     // Gibt die Wichtung der Primzahl im Sieb zurck wenn diese als dirty
     // gekennzeichnet ist, d.h. auf ein Sieben mit dieser Zahl verzichtet
     // wird; dieser Wert wird dann fr alle Elemente des Siebs einheitlich
     // als Treffer gewichtet (statt tatschlich zu sieben); blicherweise
     // ist dies ein (skalierter) Logarithmus, der in der Tabelle abgelegt
     // wurde. Diese Funktion dient dazu, den Zugriff read-only zu kapseln.
     return log_PrimeNumbers_dirty[i];
   }
private:
  static void IncreaseDirtyness(void)
   {
     // Threshold absenken und Index fr erste zu siebende Zahl inkrementieren
     if (_sieben_ab_Primzahlindex>(StaticFactorbase::Size()>>2)) return; // das geht leider nicht...
     log_Threshold-=PrimzahlSiebWichtung_dirty(_sieben_ab_Primzahlindex);
     ++_sieben_ab_Primzahlindex;
     cout << "Dirtyness now " << _sieben_ab_Primzahlindex << endl;
   }

  static void DecreaseDirtyness(void)
   {
     // Threshold hochsetzen und Index fr erste zu siebende Zahl dekrementieren
     if (_sieben_ab_Primzahlindex<=PrimzahlStartIndex) return; // das geht leider/wollen wir nicht...
     log_Threshold+=PrimzahlSiebWichtung_dirty(--_sieben_ab_Primzahlindex);
     cout << "Dirtyness now " << _sieben_ab_Primzahlindex << endl;
   }

  static void RequestDirtyness(const signed int Level)
   {
     DirtynessPegel+=Level;
     if (DirtynessPegel>1000000)
      {
        DirtynessPegel=0;
        IncreaseDirtyness();
      } else
     if (DirtynessPegel<-1000000)
      {
        DirtynessPegel=0;
        DecreaseDirtyness();
      }
   }

 public:
  static void compute_SieveThreshold(void);
  static const TSieveElement GetLogThreshold(void) { return log_Threshold; }
  static const int sieben_ab_Primzahlindex(void) { return _sieben_ab_Primzahlindex; }
  static const bool Hit_after_DirtyCorrection(const signed int Siebpos, TSieveElement Value);

  static inline void RequestMoreDirtyness(const signed int Level = 1000)
   {
     RequestDirtyness(Level);
   }
  static inline void RequestLessDirtyness(const signed int Level = 50)
   {
     RequestDirtyness(-Level);
   }

  friend void compute_StaticFactorbase();
  friend void initialize_Sieve(void);
};


// helper class to encapsulate private data
class CSieveStaticFactors
{
 private:
  static int Local_Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];
  static void init_LocalDeltas();
  static void do_sieving_StaticFactors();
  friend int main(const int argc, const char* const argv[]);
};


void do_scanning_Sieve(void);

#include <vector>
void GetStaticFactorHits(std::vector<unsigned int> &Factors, const signed int Siebpos);


#endif /* IS_SERVER not defined */

#endif /* SIEVING_HEADER_ */
