/*! @file
 * @brief
 * mpqs sieving stuff; module that is compiled separately
 */

#include "Sieving.H"
#include "qsieve.H"
#include "modulo.H"
using namespace numtheory;


int LogicalSieveSize = 1000000; // das zu jedem Polynom gehrige Siebintervall ist [-LogicalSieveSize,LogicalSieveSize]
/*
   Constraint for AMD-3DNow! specific code: abs(LogicalSieveSize)<=2^23 (because of float precision)

   LogicalSieveSize ist ein im Unterschied zu "PhysicalSieveSize" ein
   logischer Wert, der weniger stark von der verwendeten Hardware abhngig
   ist. Die logische Intervallgre ist vor allem abhngig von der Gre der
   zu faktorisierenden Zahl zu whlen.  Der hier festgelegte Wert ist also
   lediglich ein Defaultwert, der von der Konfiguration vor Beginn eines
   konkreten Faktorisierungslaufs (zur Laufzeit) modifiziert werden kann
   (und soll).
   Bemerkung: In multiple-polynomial-quadratic-sieve-Varianten sollte man
   mglichst kleine Siebintervalle whlen, da die relative Trefferquote von
   Relationen mit der Intervallgre abnimmt. (In 10 kleinen Intervallen ist
   die Chance, eine Relation zu finden, hher als in einem 10mal so groen
   Intervall).
   Gegen kleine Intervalle spricht jedoch, dass Kosten fr das Wechseln der
   Polynome, Neuinitialisierungen und Wurzelberechnungen anfallen.
*/


int SieveOffset = -LogicalSieveSize;
/* Im allgemeinen wird das fr ein Polynom zu siebende Siebintervall 
   grer sein als die (physikalisch optimale) Siebgre.
   Das Siebintervall ist daher in Teilstcken zu sieben. Hierfr wird der
   Sieboffset bentigt, der den Startpunkt des jeweiligen Teilsiebs angibt.
 */


class StaticRelations
{
 public:
  static void insert(CRelation *GL, const bool do_multi_combine_init=true);
};

TSieveElement SieveArray_[PhysicalSieveSize+64+8] __attribute__ ((aligned (64))); // Speicher fr das Sieb

TSieveElement log_Primzahl_of_PrimePowers[StaticFactorbase::max_additional_Powers];
/* fr weitere Primzahlpotenzen stehen hier die konvertierten Logarithmen der Primzahlen */


TSieveElement SieveControl::log_PrimeNumbers[StaticFactorbase::MaxSize];
TSieveElement SieveControl::log_PrimeNumbers_dirty[StaticFactorbase::MaxSize];
int SieveControl::_sieben_ab_Primzahlindex = SieveControl::PrimzahlStartIndex;
signed int SieveControl::DirtynessPegel = 0;


const bool SieveControl::Hit_after_DirtyCorrection(const signed int Siebpos, TSieveElement Value)
{
  // An dieser Stelle im Sieb wurde ein Treffer gefunden;
  // aber wegen der Dirty-Siebfaktoren, die fr alle Siebelemente als Treffer gezhlt werden,
  // knnte es tatschlich gar kein Treffer sein.
  // Wir mssen daher den Dirty-Siebwert um die nicht-teilenden Dirty-Faktorbasiselemente korrigieren.
  // Dieser Korrekturwert wird hier bestimmt (allerdings mit shortcut).

  // Wenn diese Routine aufgerufen wird, ist mglicherweise ein Siebtreffer gelandet worden;
  // Da es sich um ein Regelungssystem handelt, ist dies (zwar kein perfekter, aber)
  // ein sinnvoller Indikator, um nun wieder etwas weniger Dirtyness zu verlangen,
  // wenn doch kein Siebtreffer vorlag...

  for (int i=_sieben_ab_Primzahlindex-1; i>=PrimzahlStartIndex; --i)
   {
     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[i]);
     // same as: Siebpos%PrimeNumbers[i]; if (d<0) d+=PrimeNumbers[i];
     // Bemerkung: %-operation does *not* eliminate the sign!
     //            ugly: -1=1 (mod 2), but (-1)%2=-1, 1%2=1
     // -> so we have to eliminate the sign!!!
     const bool Treffer = (d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]);
     if (Treffer)
      {
        // Value verkleinert sich
        Value+=PrimzahlSiebWichtung_dirty(i); // dirty-Wichtung wieder wegnehmen
	Value-=PrimzahlSiebWichtung(i); // und durch die echte Wichtung ersetzen
      }
     else
      {
        // hat nicht getroffen:
        // Value vergert sich
        Value+=PrimzahlSiebWichtung_dirty(i); // dirty-Wichtung wegnehmen, da kein Treffer vorlag

        // shortcut "return false" on value>=0 doesn't work anymore, since it's not monotone increasing
        //  2004-03-12: maybe we should revert to the old behaviour?
         // the more dirty the sieving is -> the more values to analyse
         // the more expensive divisions we need to take and we cannot shortcut! :-(
      }
   }

  if (Value>=0)
   {
     SieveControl::RequestLessDirtyness(); // "gegenregeln"
     return false; // Threshold berschritten, kein Siebtreffer
   }
  return true; // Threshold nicht berschritten -> Siebtreffer
}


TSieveElement SieveControl::log_Threshold = 0; // wird aber erst durch "Grundinitialisierung_des_Siebs" initialisiert

void SieveControl::compute_SieveThreshold(void)
{
  // Hier wird der Threshold (berschlagsmig) ermittelt
  
  // Voraussetzungen zum Aufruf der Initialisierung:
  // 1. Logische Siebintervallgre LogicalSieveSize (fr Intervall [-LogicalSieveSize,LogicalSieveSize])
  // 2. kN (die mit Vorfaktor versehene zu faktorisierende Zahl)
  // 3. die statische Faktorbasis
  // mssen bereits bekannt und berechnet sein.
 
  // logarithmischen Schwellwert kalkulieren:
  mpz_t x; mpz_init(x);
  mpz_div_ui(x,kN,2); mpz_sqrt(x,kN); mpz_mul_ui(x,x,LogicalSieveSize);
  double d = log(fabs(mpz_get_d(x))) - (Factor_Threshold*log(biggest_Prime_in_Factorbase));
  mpz_clear(x);
  if (PrimeNumbers[1]==2) d-=log(2); // Falls 2 in der Faktorbasis, dann teilt 2 alle Siebeintrge
  d*=log_Multiplikator;
  cout << "Threshold (initial sieve entries) for detecting relations set to " << d << endl;
  log_Threshold=static_cast<TSieveElement>(floor(d));
  if (log_Threshold!=floor(d)) 
   {
     cerr << "With the compiled-in types and constants this" << endl;
     cerr << "factorization cannot be done, sorry." << endl; 
     cerr << "TSieveElement (compile-time type) needs to be increased to fit the threshold!" << endl;
     cerr << "Please modify log_Multiplikator and/or TSieveElement" << endl;
     cerr << "and compile the program again." << endl;
     exit(1);
   }

#if TSIEBELEMENTSIZE > 0
  // sanity check...
  if (sizeof(TSieveElement)!=TSIEBELEMENTSIZE)
   {
     cerr << "Compiled types and values for " << endl
          << "TSieveElement (Size=" << sizeof(TSieveElement) << ")"
          << " and TSIEBELEMENTSIZE (=" << TSIEBELEMENTSIZE << ")" << endl
          << "do not match. Please correct them!" << endl;
     exit(1);
   }
#endif
}




#ifdef ASM386
#define SIEBASM386

unsigned int clobbered_int; // dummy

#if TSIEBELEMENTSIZE == 1
#define asm_sieb_init() \
  asm volatile ( \
       "movb %%al,%%ah\n\t" \
       "pushw %%ax\n\t" \
       "shll $16,%%eax\n\t" \
       "popw %%ax\n\t" \
       "shrl $2,%%ecx\n\t" \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "cc", "memory");

#ifdef ASMATHLON
// ATHLON-optimized code
#ifdef DEBUG
#warning "using ATHLON version for macro: asm_search_sieb"
#endif
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $7,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "jns 0b \n\t" \
       "js 4f # Treffer gefunden \n\t" \
       "1: # Treffer suchen (big steps) loop \n\t" \
       "prefetch -256(%[sieb],%[pos]) # tuning cache!! \n\t" \
       "subl $64,%[pos] \n\t" \
       "movq  (%[sieb],%[pos]),%%mm0 \n\t" \
       "movq 8(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 16(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 24(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 32(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 40(%[sieb],%[pos]),%%mm1 \n\t" \
       "por 48(%[sieb],%[pos]),%%mm0 \n\t" \
       "por 56(%[sieb],%[pos]),%%mm1 \n\t" \
       "por %%mm1,%%mm0 \n\t" \
       "pmovmskb %%mm0,%%eax # highest bits of packed bytes to %%ah \n\t" \
       "testl %%eax,%%eax \n\t" \
       "jz 1b \n\t" \
       "femms # clear MMX state \n\t" \
       "2: # Treffer suchen (little steps) init \n\t" \
       "addl $64,%[pos] \n\t" \
       "3: # Treffer suchen (little steps) loop \n\t" \
       "subl $4,%[pos] \n\t" \
       "testl $0x80808080,(%[sieb],%[pos]) \n\t" \
       "jz 3b \n\t" \
       "movl (%[sieb],%[pos]),%%eax \n\t" \
       "addl $3,%[pos] \n\t" \
       "testl $0x80000000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x00800000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x00008000,%%eax \n\t" \
       "jnz 4f \n\t" \
       "decl %[pos] \n\t" \
       "4: # Treffer gefunden" \
       : [pos] "=q" (offset) : "[pos]" (offset), [sieb] "r" (SieveArray) : "cc", "eax", "mm0", "mm1");
#else
// optimized i386 code
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $3,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "jns 0b \n\t" \
       "js 2f # Treffer gefunden \n\t" \
       "1: # Treffer suchen \n\t" \
       "subl $4,%[pos] \n\t" \
       "testl $0x80808080,(%[sieb],%[pos]) \n\t" \
       "jz 1b \n\t" \
       "addl $3,%[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "testb $0x80,(%[sieb],%[pos]) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#endif
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmpl %[disp1],%[disp0]\n\t" \
       "jle 3f\n\t" \
       "xchgl %[disp1],%[disp0]\n" \
       "3:\n\t" \
       "testl %[disp0],%[disp0]\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subb %[val],(%[feld],%[disp1])\n\t" \
       "subb %[val],(%[feld],%[disp0])\n\t" \
       "subl %[step],%[disp1]\n\t" \
       "subl %[step],%[disp0]\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "testl %[disp1],%[disp1]\n\t" \
       "jl 2f\n\t" \
       "subb %[val],(%[feld],%[disp1])\n" \
       "2:\n" \
       : [disp0] "=q" (clobbered_int), [disp1] "=q" (clobbered_int) : [feld] "r" (SieveArray), [step] "g" (P), [val] "r" (lp), "[disp0]" (d0), "[disp1]" (d1) : "cc");

#elif TSIEBELEMENTSIZE == 2
#define asm_sieb_init() \
  asm volatile ( \
       "pushw %%ax\n\t" \
       "shll $16,%%eax\n\t" \
       "popw %%ax\n\t" \
       "shrl $1,%%ecx\n\t" \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "cc", "memory");
#define asm_search_sieb(offset) \
  asm volatile ( \
       "0: # Alignment suchen \n\t" \
       "testl $1,%[pos] \n\t" \
       "jz 1f # is aligned \n\t" \
       "decl %[pos] \n\t" \
       "testw $0x8000,(%[sieb],%[pos],2) \n\t" \
       "js 2f # Treffer gefunden \n\t" \
       "1: # Treffer suchen \n\t" \
       "subl $2,%[pos] \n\t" \
       "testl $0x80008000,(%[sieb],%[pos],2) \n\t" \
       "jz 1b \n\t" \
       "incl %[pos] \n\t" \
       "testw $0x8000,(%[sieb],%[pos],2) \n\t" \
       "js 2f \n\t" \
       "decl %[pos] \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subw %4,(%2,%1,2)\n\t" \
       "subw %4,(%2,%0,2)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subw %4,(%2,%1,2)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (SieveArray), "g" (P), "q" (lp), "0" (d0), "1" (d1) : "cc");
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (SieveArray), "g" (P), "q" (lp) : "cc");
#elif TSIEBELEMENTSIZE == 4
#define asm_sieb_init() \
  asm volatile ( \
       "cld\n\t" \
       "rep\n\t" \
       "stosl" \
       : "=c" (clobbered_int), "=D" (clobbered_int) : "a" (SieveControl::log_Threshold), "D" (SieveArray), "c" (PhysicalSieveSize) : "memory");
#define asm_search_sieb(offset) \
  asm volatile ( \
       "1: # Treffer suchen \n\t" \
       "decl %[pos] \n\t" \
       "testl $0x80008000,(%[sieb],%[pos],4) \n\t" \
       "jz 1b \n\t" \
       "2: # Treffer gefunden" \
       : [pos] "=r" (offset) : "[pos]" (offset), [sieb] "q" (SieveArray) : "cc");
#define asm_sieb(lp, d0, d1, P) \
  asm volatile ( \
       "cmp %1,%0\n\t" \
       "jle 3f\n\t" \
       "xchgl %1,%0\n" \
       "3:\n\t" \
       "test %0,%0\n\t" \
       "jl 1f\n" \
       "0:\n\t" \
       "subl %4,(%2,%1,4)\n\t" \
       "subl %4,(%2,%0,4)\n\t" \
       "subl %3,%1\n\t" \
       "subl %3,%0\n\t" \
       "jns 0b\n" \
       "1:\n\t" \
       "test %1,%1\n\t" \
       "jl 2f\n\t" \
       "subl %4,(%2,%1,4)\n" \
       "2:\n" \
       : "=r" (d0), "=r" (d1) : "q" (SieveArray), "g" (P), "q" (lp), "0" (d0), "1" (d1) : "cc");
// fr neueren gcc funktioniert auch  : "+r" (d0), "+r" (d1) : "q" (SieveArray), "g" (P), "q" (lp) : "cc");
#else
#undef SIEBASM386
#warning "Kein handoptimierter Asssemblerkode fr Aussieben mit TSieveElement!"
#endif
#endif


void initialize_Sieve(void)
{
  //cout << "Initializing sieve from " << SieveOffset
  //     << " to " << SieveOffset+PhysicalSieveSize << endl;

 SieveControl::RequestMoreDirtyness();

#ifdef SIEBASM386
  asm_sieb_init(); // siehe das entsprechende Assembler-Makro
#else
  register TSieveElement w = SieveControl::GetLogThreshold();
  /* log_Threshold bezeichnet den Schwellwert, der zum Erkennen der Relation
     unterschritten werden mu. Es ist besser, den Threshold bereits bei der
     Initialisierung zu bercksichtigen (d.h. abzuziehen), da beim Auslesen
     des Siebs dann nur auf negative Werte getestet werden braucht. */
  
  // Initialisierung des Siebs mit dem berschlagsmigen Schwellwert
  register int i=PhysicalSieveSize; do SieveArray[--i]=w; while (i);
#endif

  // nachfolgender Block dient (falls aktiviert) nur noch Validierungszwecken:
  // enable this block to compare the polynomial value and its logval against they actually used threshold
#if 0
  {
    mpz_t x,y;
    mpz_init(x); mpz_init(y);
    Polynom.get_values(SieveOffset,x,y);
    TSieveElement w_test = static_cast<TSieveElement>(log_Multiplikator*log(fabs(mpz_get_d(y))));
    TSieveElement FB_Threshold
      = static_cast<TSieveElement>(log_Multiplikator*Factor_Threshold*log(biggest_Prime_in_Factorbase));
    w_test-=FB_Threshold;
    if (PrimeNumbers[1]==2) w_test-=static_cast<TSieveElement>(log_Multiplikator*log(2));
 #if TSIEBELEMENTSIZE == 1
    // chars are not printed as numbers, so we cast them:
    cout << "value " << y << " -> logval: " << static_cast<signed int>(w_test)
         << " <-> threshold: " << static_cast<signed int>(SieveControl::GetLogThreshold()) << endl;
 #else
    cout << "value " << y << " -> logval: " << w_test << " <-> threshold: " << SieveControl::GetLogThreshold() << endl;
 #endif
    mpz_clear(x); mpz_clear(y);
  }
#endif
}


int CSieveStaticFactors::Local_Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2];


void CSieveStaticFactors::init_LocalDeltas()
{
  int nr=StaticFactorbase::Size()-1;
  // groe Primzahlen (maximal 1 Siebtreffer pro Delta) 
  for (;; --nr) 
    {
      register int Primzahl=PrimeNumbers[nr];

      if (Primzahl<PhysicalSieveSize) break;

      // d0/d1-Ermittlung
      register int d0,d1;
      {
        register int h=normalized_signed_mod(SieveOffset,Primzahl); // damned signed %-Operation...
	d0=Delta_of_PrimeNumbers[nr][0]-h;
	d1=Delta_of_PrimeNumbers[nr][1]-h;
	if (d1<0)
          d1+=Primzahl;
	if (d0<0)
          d0+=Primzahl;
      }
     Local_Delta_of_PrimeNumbers[nr][0]=d0;
     Local_Delta_of_PrimeNumbers[nr][1]=d1;
    }
}

void CSieveStaticFactors::do_sieving_StaticFactors()
{
  //cout << "Aussieben mit Primzahlen der statischen FB: " << SieveOffset<< endl;
  
  /* wir sieben erst ab einer greren Primzahl aus der Faktorbasis,
     da wir sowieso unsauber und schnell sieben...
     ( je kleiner die Primzahl, um langsamer ist das Sieben mit dieser
     Zahl; und bei der gewhlten Implementierung ohne Bercksichtigung
     von Primzahlpotenzen ist auch der zustzlich gemachte Fehler
     vernachlssigbar und wird zudem ber Factor_Threshold kompensiert.
     Aber Vorsicht: Shortcut-Konstante nicht zu hoch ansetzen, sonst fallen
     zu viele Relationen raus!
     Achtung: als Startwert mindestens 2 whlen, damit -1 und 2 ignoriert
     werden knnen. Dann gibt es auch 2 Deltawerte. )
     -> relevant fr Schleife mit den kleinen Primzahlen
  */

  // standard code, but uses also various assembler optimizations

  int nr=StaticFactorbase::Size()-1;
  // groe Primzahlen (maximal 1 Siebtreffer pro Delta) 
  for (;; --nr) 
    {
      register int Primzahl=PrimeNumbers[nr];

      if (Primzahl<PhysicalSieveSize) break;

      // d0/d1-Ermittlung
      register int d0,d1;
      
      d0=Local_Delta_of_PrimeNumbers[nr][0];
      d1=Local_Delta_of_PrimeNumbers[nr][1];

      register TSieveElement log_Primzahl = SieveControl::PrimzahlSiebWichtung(nr);
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      SieveArray[d0<PhysicalSieveSize ? d0 : PhysicalSieveSize]-=log_Primzahl;
      SieveArray[d1<PhysicalSieveSize ? d1 : PhysicalSieveSize]-=log_Primzahl;

      d0-=PhysicalSieveSize; if (d0<0) d0+=Primzahl;
      d1-=PhysicalSieveSize; if (d1<0) d1+=Primzahl;
      Local_Delta_of_PrimeNumbers[nr][0]=d0;
      Local_Delta_of_PrimeNumbers[nr][1]=d1;
    }

  // kleine Primzahlen (mehrere Siebtreffer pro Delta)
  const int NonDirtyIndex = SieveControl::sieben_ab_Primzahlindex();
  for (; nr>=NonDirtyIndex; --nr) 
    {
      register int Primzahl = PrimeNumbers[nr];
      //cout << "Quadratisches Sieb mit " << Primzahl << endl;
      
      // d0/d1-Ermittlung
      register int d0,d1;
      {
	register int h=normalized_signed_mod(SieveOffset+PhysicalSieveSize,Primzahl); // damned signed %-Operation...
	d0=Delta_of_PrimeNumbers[nr][0]-h;
	d1=Delta_of_PrimeNumbers[nr][1]-h;
	if (d1>=0) d1-=Primzahl;
	if (d0>=0) d0-=Primzahl;
      }
      
      register TSieveElement log_Primzahl = SieveControl::PrimzahlSiebWichtung(nr);
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      d0+=PhysicalSieveSize;
      d1+=PhysicalSieveSize;

      //if (d0==d1) cout << Primzahl << " teilt Vorfaktor!" << endl;
      
#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, Primzahl);
#else 
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so dass d0<d1
      while (d0>=0)
	{
	  SieveArray[d0]-=log_Primzahl; d0-=Primzahl;
	  SieveArray[d1]-=log_Primzahl; d1-=Primzahl;
	}
      if (d1>=0) SieveArray[d1]-=log_Primzahl;
#endif
    }

  // hhere Primzahlpotenzen
  for (nr=0; nr<NumberOf_more_PrimePowers; nr++)
    {
      register int PPotenz = PrimePowers[nr];

      // d0/d1-Ermittlung
      register int d0,d1;
      {
	register int h=normalized_signed_mod(SieveOffset+PhysicalSieveSize,PPotenz); // damned signed %-Operation...
	d0=Delta_of_PrimePowers[nr][0]-h;
	d1=Delta_of_PrimePowers[nr][1]-h;
	if (d1>=0) d1-=PPotenz;
	if (d0>=0) d0-=PPotenz;
      }
      
      register TSieveElement log_Primzahl = log_Primzahl_of_PrimePowers[nr];
      /* d.h. log_Primzahl erhlt den Wert (TSieveElement) (log_Multiplikator*log(Primzahl)) */
      
      // beachte: da mit den Primzahlen selbst auch gesiebt wird, darf nicht log(Primzahlpotenz), sondern
      // nur log(Primzahl) abgezogen werden!
      // Analogie: Wenn man schon durch p geteilt hat, mu man nocheinmal durch p teilen,
      //           um insgesamt durch p^2 zu teilen!

      d0+=PhysicalSieveSize;
      d1+=PhysicalSieveSize;

#ifdef SIEBASM386
      asm_sieb(log_Primzahl, d0, d1, PPotenz);
#else
      if (d0>d1) { int h=d0; d0=d1; d1=h; } // ggf. tauschen, so dass d0<d1
      while (d0>=0)
	{
	  SieveArray[d0]-=log_Primzahl; d0-=PPotenz;
	  SieveArray[d1]-=log_Primzahl; d1-=PPotenz;
	}
      if (d1>=0) SieveArray[d1]-=log_Primzahl;      
#endif
    }

}


void do_scanning_Sieve(void)
{
  //cout << "Auslesen des (gesiebten) Siebintervalls" << endl;

  // da nur ein Bruchteil der Siebelemente Treffer sind, sollte man das Suchen beschleunigen!
  // Constraint: the 64 positions below the sieve (SieveArray[-64]..SieveArray[-1]) are already set to -1.
  register int d=PhysicalSieveSize;
  do
   {
     // todo: hier bietet sich in assembler doch frmlich an...
#ifdef SIEBASM386
     asm_search_sieb(d); // nach Siebtreffer suchen...
     //if (d>=0 && SieveArray[d]<0) cout << "Siebtreffer fr " << d << endl;
     //else cout << "??? fr " << d << endl; // d==-1 is okay...
#else
     while (SieveArray[--d]>=0);; // nach Siebtreffer suchen...
#endif
     if (d>=0) // gefundener Treffer *nicht* Dummytreffer?
      {
        if (SieveControl::Hit_after_DirtyCorrection(SieveOffset+d,SieveArray[d]))
          StaticRelations::insert(new CRelation(SieveOffset+d)); // dann einfgen
      }
   } while (d>0); // solange noch regulre Treffer mglich: weitersuchen
// Werteverteilung_im_Sieb_ausgeben();
}


#ifdef ASM_P3_SSE /* special code using SSE instructions (P3) */

void GetStaticFactorHits(std::vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )

  // (Pentium3) SSE specific code

  #warning "new Pentium3/SSE specific code"
  typedef int v4sf __attribute__ ((mode(V4SF)));
  register v4sf XMM6 asm("xmm6"); // XMM6 is now bound to %xmm6 (intention: signedness correction)
  register v4sf XMM7 asm("xmm7"); // XMM7 is now bound to %xmm7

   // refer GCC manual, "Specifying registers for local variables";
   // but remember: that does not necessarily mean, that %xmm7 cannot be
   // used for other variables as well! (It is only guaranteed, that if we access
   // XMM7, this access is synonym to accessing %xmm7; but accessing %xmm7 is not necessarily
   // synonym to accessing XMM7 (because if the compiler assumes, that XMM7 is dead, %xmm7 could
   // be used for other variables; and the compiler does *not* know, that we want
   // XMM7 to be alive by accessing %xmm7!).

  asm ( \
       "emms                       # enter mmx\n\t" \
       "movd %[Sieboffs],%%mm0   # -      mm0=(0,Sieboffs)\n\t" \
       "xorps %[_xmm6],%[_xmm6]    # L2AM   clear mm6 \n\t" \
       "punpckldq %%mm0,%%mm0  # L2AM   (0,Sieboffs) -> (Sieboffs,Sieboffs) \n\t" \
       "cvtpi2ps %%mm0,%[_xmm7]  # packed int to float \n\t" \
       "shufps $0x44,%[_xmm7],%[_xmm7] \n\t" \
       "cmpnleps %[_xmm7],%[_xmm6] # L2AM   set bits if signed, clear bits if unsigned" \
       : [_xmm7] "=x" (XMM7), [_xmm6] "=x" (XMM6) : [Sieboffs] "g" (Siebpos) : "mm0");

  // PrimeNumbers[0] und PrimeNumbers[1] werden ausgelassen (da sie gesondert behandelt werden mssen)
  // PrimeNumbers[2] und PrimeNumbers[3] werden hier gesondert behandelt (damit der restliche Anzahl durch 4 teilbar ist)
  {
     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[2]);
     if (d==Delta_of_PrimeNumbers[2][0] || d==Delta_of_PrimeNumbers[2][1]) Factors.push_back(2);
     d = normalized_signed_mod(Siebpos,PrimeNumbers[3]);
     if (d==Delta_of_PrimeNumbers[3][0] || d==Delta_of_PrimeNumbers[3][1]) Factors.push_back(3);
  }
  
  for (int nr=4; nr<StaticFactorbase::Size(); nr+=4) 
    {
      register unsigned int pflags, pflags2;
      asm ( \
       "movq %[divisors](,%[i],4),%%mm4   # L2AMS  get y : x (as packed int) \n\t" \
       "movq 8+%[divisors](,%[i],4),%%mm5   # L2AMS  get y : x (as packed int) \n\t" \
       "cvtpi2ps %%mm4,%%xmm0 # convert packed int to float \n\t" \
       "cvtpi2ps %%mm5,%%xmm1 # convert packed int to float \n\t" \
       "movlhps %%xmm1,%%xmm0 \n\t" \
       "movaps %[_xmm7],%%xmm2 \n\t" \
       "movaps %%xmm0,%%xmm1 \n\t" \
       "divps %%xmm0,%%xmm2 \n\t" \
       "movaps %%xmm1,%%xmm3 \n\t" \
       "movhlps %%xmm2,%%xmm4 \n\t" \
       "cvttps2pi %%xmm2,%%mm0 \n\t" \
       "cvttps2pi %%xmm4,%%mm1 \n\t" \
       "movaps %[_xmm7],%%xmm2 \n\t" \
       "cvtpi2ps %%mm0,%%xmm0 \n\t" \
       "cvtpi2ps %%mm1,%%xmm4 \n\t" \
       "movlhps %%xmm4,%%xmm0 \n\t" \
       "mulps %%xmm1,%%xmm0 \n\t" \
       "subps %%xmm0,%%xmm2 \n\t" \
       "andps %[_xmm6],%%xmm1      # correct sign (part 1) \n\t" \
       "addps %%xmm1,%%xmm2        # correct sign (part 2) \n\t" \
       "movaps %%xmm2,%%xmm0 \n\t" \
       "cmpneqps %%xmm3,%%xmm0      #  assure modp=0 instead of modp=p \n\t" \
       "andps %%xmm0,%%xmm2   \n\t" \
       "movhlps %%xmm2,%%xmm4 \n\t" \
       "cvttps2pi %%xmm2,%%mm0 \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd %[Deltas](,%[i],8),%%mm1   # L2AM \n\t" \
       "pcmpeqd 8+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags]   # L6+ (VectorPath)  into bytemask \n\t" \
       "cvttps2pi %%xmm4,%%mm0 \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd 16+%[Deltas](,%[i],8),%%mm1 # L2AM \n\t" \
       "pcmpeqd 24+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags2]  # L6+ (VectorPath)  into bytemask" \
       : [pflags] "=r" (pflags), [pflags2] "=r" (pflags2)
       : [divisors] "o" (PrimeNumbers[0]), [Deltas] "o" (Delta_of_PrimeNumbers[0][0]),
         [_xmm7] "x" (XMM7), [_xmm6] "x" (XMM6), [i] "r" (nr)
       : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "xmm0", "xmm1", "xmm2", "xmm3", "xmm4");

      //cout << "Siebpos: " << Siebpos << ": " << PackedResult[0] << "," << PackedResult[1] << endl;
 
#if defined(DEBUG)
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr]);
         if ( (d==Delta_of_PrimeNumbers[nr][0]) != ((pflags&0x03)==0x03) ) { MARK; cerr << nr << "!!" << (pflags&0x03) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr][1]) != ((pflags&0x0c)==0x0c) ) { MARK; cerr << nr << "!!" << (pflags&0x0c) << endl; exit(7); }
       }
#endif
      if ( pflags & 0x0f ) Factors.push_back(nr);

#if defined(DEBUG)
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+1]);
         if ( (d==Delta_of_PrimeNumbers[nr+1][0]) != ((pflags&0x30)==0x30) ) { MARK; cerr << nr+1 << "!!" << (pflags&0x30) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+1][1]) != ((pflags&0xc0)==0xc0) ) { MARK; cerr << nr+1 << "!!" << (pflags&0xc0) << endl; exit(7); }
       }
#endif
      if ( pflags & 0xf0 ) Factors.push_back(nr+1);

#if defined(DEBUG)
       {
         // check pflags2 for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+2]);
         if ( (d==Delta_of_PrimeNumbers[nr+2][0]) != ((pflags2&0x03)==0x03) ) { MARK; cerr << nr+2 << "!!" << (pflags2&0x03) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+2][1]) != ((pflags2&0x0c)==0x0c) ) { MARK; cerr << nr+2 << "!!" << (pflags2&0x0c) << endl; exit(7); }
       }
#endif
      if ( pflags2 & 0x0f ) Factors.push_back(nr+2);

#if defined(DEBUG)
       {
         // check pflags2 for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+3]);
         if ( (d==Delta_of_PrimeNumbers[nr+3][0]) != ((pflags2&0x30)==0x30) ) { MARK; cerr << nr+3 << "!!" << (pflags2&0x30) << endl; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+3][1]) != ((pflags2&0xc0)==0xc0) ) { MARK; cerr << nr+3 << "!!" << (pflags2&0xc0) << endl; exit(7); }
       }
#endif
      if ( pflags2 & 0xf0 ) Factors.push_back(nr+3);
    }
  asm("emms                      # exit mmx");
}
// end of SSE optimized version


#elif defined(ASMATHLON) /* special code for AMD 3DNow! */

void GetStaticFactorHits(std::vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )

  // Athlon 3DNow! specific code

  #warning "new AMD 3DNow! specific code"
  typedef int v2sf __attribute__ ((mode(V2SF)));
  register v2sf MM7 asm("mm7"); // MM7 is now bound to %mm7
  register v2sf MM6 asm("mm6"); // MM6 is now bound to %mm6 (intention: signedness correction)

   // refer GCC manual, "Specifying registers for local variables";
   // but remember: that does not necessarily mean, that %mm7 cannot be
   // used for other variables as well! (It is only guaranteed, that if we access
   // MM7, this access is synonym to accessing %mm7; but accessing %mm7 is not necessarily
   // synonym to accessing MM7 (because if the compiler assumes, that MM7 is dead, %mm7 could
   // be used for other variables; and the compiler does *not* know, that we want
   // MM7 to be alive by accessing %mm7!).

  asm ( \
       "femms                      # L2AMS  fast enter mmx\n\t" \
       "movd %[Sieboffs],%[_mm7]   # -      MM7=(0,Sieboffs)\n\t" \
       "punpckldq %[_mm7],%[_mm7]  # L2AM   (0,Sieboffs) -> (Sieboffs,Sieboffs) \n\t" \
       "movq %[_mm7],%%mm0         # L2AM \n\t" \
       "pi2fd %[_mm7],%[_mm7]      # L4A    int to float \n\t" \
       "pxor %[_mm6],%[_mm6]       # L2AM   clear mm6 \n\t" \
       "pcmpgtd %%mm0,%%mm6        # L2AM   set bits if signed, clear bits if unsigned" \
       : [_mm7] "=y" (MM7), [_mm6] "=y" (MM6) : [Sieboffs] "g" (Siebpos) : "mm0");

  // PrimeNumbers[0] und PrimeNumbers[1] werden ausgelassen (da sie gesondert behandelt werden mssen)
  for (int nr=2; nr<StaticFactorbase::Size(); nr+=2) 
    {
      register unsigned int pflags;
      asm ( \
       "movq %[divisors](,%[i],4),%%mm4   # L2AMS  get y : x (as packed int)\n\t" \
       "pi2fd %%mm4,%%mm0          # L4A    get y : x (as packed float)\n\t" \
       "movq %%mm4,%%mm5           # L2AMS \n\t" \
       "pfrcp %%mm0,%%mm1          # L3M    1/y : 1/x (approx.) \n\t" \
       "movq %%mm0,%%mm2           # L2AM   y : x -> mm0 \n\t" \
       "punpckhdq %%mm0,%%mm0      # L2AM   y : y \n\t" \
       "movq %%mm2,%%mm3           # L2AM   y : x -> mm3 \n\t" \
       "pfrcp %%mm0,%%mm0          # L3M    1/y : 1/y (approx.) \n\t" \
       "punpckldq %%mm0,%%mm1      # L2AM   1/y : 1/x (approx.) \n\t" \
       "pfrcpit1 %%mm1,%%mm2       # L4M    1/y : 1/x (refine) \n\t" \
       "pfrcpit2 %%mm1,%%mm2       # L4M    1/y : 1/x (final) \n\t" \
       "pand %%mm6,%%mm4           # L2AM   correct sign (part 1) \n\t" \
       "pfmul %[_mm7],%%mm2        # L4M    z/y : w/x \n\t" \
       "pf2id %%mm2,%%mm1          # L4A    packed float to int \n\t" \
       "pi2fd %%mm1,%%mm1          # L4A    and back (->floor) \n\t" \
       "pfmul %%mm3,%%mm1          # L4M    multiply \n\t" \
       "pfsubr %[_mm7],%%mm1       # L4A    subtract \n\t" \
       "pf2id %%mm1,%%mm0          # L4A    ->modulo \n\t" \
       "paddd %%mm4,%%mm0          # L2AM   correct sign (part 2) \n\t" \
       "movq  %%mm0,%%mm4          # L2AM \n\t" \
       "pcmpeqd %%mm5,%%mm0        # L2AM  assure modp=0 instead of modp=p \n\t" \
       "pandn %%mm4,%%mm0          # L2AM \n\t" \
       "pshufw $0x44,%%mm0,%%mm1   # L2AM \n\t" \
       "pshufw $0xee,%%mm0,%%mm2   # L2AM \n\t" \
       "pcmpeqd %[Deltas](,%[i],8),%%mm1   # L2AM \n\t" \
       "pcmpeqd 8+%[Deltas](,%[i],8),%%mm2 # L2AM \n\t" \
       "packssdw %%mm2,%%mm1       # L2AM  pack comparison results together \n\t" \
       "pmovmskb %%mm1,%[pflags]   # L6+ (VectorPath)  into bytemask" \
       : [pflags] "=r" (pflags)
       : [divisors] "o" (PrimeNumbers[0]), [Deltas] "o" (Delta_of_PrimeNumbers[0][0]),
         [_mm7] "y" (MM7), [_mm6] "y" (MM6), [i] "r" (nr)
       : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5");

      //cout << "Siebpos: " << Siebpos << ": " << PackedResult[0] << "," << PackedResult[1] << endl;
 
#ifdef DEBUG
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr]);
         if ( (d==Delta_of_PrimeNumbers[nr][0]) != ((pflags&0x03)==0x03) ) { MARK; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr][1]) != ((pflags&0x0c)==0x0c) ) { MARK; exit(7); }
       }
#endif
      if ( pflags & 0x0f ) Factors.push_back(nr);

#ifdef DEBUG
       {
         // check pflags for correct byte mask
         const int d=normalized_signed_mod(Siebpos,PrimeNumbers[nr+1]);
         if ( (d==Delta_of_PrimeNumbers[nr+1][0]) != ((pflags&0x30)==0x30) ) { MARK; exit(7); }
         if ( (d==Delta_of_PrimeNumbers[nr+1][1]) != ((pflags&0xc0)==0xc0) ) { MARK; exit(7); }
       }
#endif
      if ( pflags & 0xf0 ) Factors.push_back(nr+1);
    }
  asm("femms                      # L2AMS  fast exit mmx");
}
// end of AMD 3DNow! optimized version

#else /* generic version */

void GetStaticFactorHits(std::vector<unsigned int> &Factors, const signed int Siebpos)
{
  // find factors in static factorbase
  // (beginnend mit der 3. Primzahl der FB (s.o) )
  for (int i=2; i<StaticFactorbase::Size(); i++)
   {
     // shortcut:
     /* keine aufwendige Division durchfhren, wenn diese
        Primzahl gar nicht ausgesiebt wurde */ 

     register int d = normalized_signed_mod(Siebpos,PrimeNumbers[i]);
     // same as: Siebpos%PrimeNumbers[i]; if (d<0) d+=PrimeNumbers[i];
     // Bemerkung: %-operation does *not* eliminate the sign!
     //            ugly: -1=1 (mod 2), but (-1)%2=-1, 1%2=1
     // -> so we have to eliminate the sign!!!
     // and don't try to use unsigned int d, this won't work!

#if 0
     // for debugging reasons:
     bool b1=(mpz_div_ui(x,y,PrimeNumbers[i])==0);
     bool b2=(d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]);
     if (b1!=b2) 
      {
        cerr << "error in shortcut: " << Siebpos << "; "
             << PrimeNumbers[i] << ": " << d << ", "
             << Delta_of_PrimeNumbers[i][0] << ", "
             << Delta_of_PrimeNumbers[i][1] << endl;
        exit(1);
      }
#endif

     if (d==Delta_of_PrimeNumbers[i][0] || d==Delta_of_PrimeNumbers[i][1]) Factors.push_back(i);
   }
}
#endif /* generic code */
