// Modulo-Operations for unsigned int
// written by Thorsten Reinecke (1999-03-12)
// last change: 2004-07-27

// some of the following functions are necessary because a*b%m will
// return false results if a*b overflows unsigned-int-range.
// these routines are okay for unsigned-int-range/2

/*! @file
 * @brief
 * This header file contains some number theoretic functions.
 *
 * The functions implemented (or declared) in this file provide support for
 * fast modulo operations (multiplication, exponentiation, computation of
 * the modular inverse, quadratic residues) and some integer functions
 * (greatest common divisor, prime numbers).
 *
 * Most of these functions contain also optimized assembler code for
 * Intel Pentium and AMD Athlon processors.
 */


#ifndef MODULO_H_INCLUDED
#define MODULO_H_INCLUDED

#include <iostream>

//! number theoretic concepts
namespace numtheory
{


/*!
 * @param x a signed integer value
 * @param m an unsigned integer value
 * @result @p x mod @p m
 * @remark This function is a replacement for x%m. It normalizes
 *         the result to a nonnegative number.
 *         (On most architectures x%m is negative, if x is negative.) 
 */
inline unsigned int normalized_signed_mod(const signed int x, const int m)
{
#if defined(ASM386_CMOV)
  // *** inline x86 assembler using conditional mov operations ***
#ifdef DEBUG
 #warning "using cmov-operation for normalized_signed_mod(const signed int,int)"
#endif
  // damned sign!!
  // x%m will not be normalized by default, but often you depend on it!
  // -1%2 -> -1 , 1%2 -> 1,
  // therefore (-1%2 != 1%2) but one surely expects -1=1 (mod 2) !!!
  register unsigned int result;
  register unsigned int clobbered;
  asm ( \
   "cdq # prepare signed division \n\t"
   "idivl %[mod] # remainder -> edx \n\t" \
   "mov %%edx,%%eax \n\t" \
   "addl %[mod],%%edx # this is a positive value \n\t" \
   "test %%eax,%%eax \n\t" \
   "cmovns %%eax,%%edx # and now it is normalized (minimal) \n\t"
   : "=a" (clobbered), "=&d" (result) : "a" (x), [mod] "g" (m) : "cc");
  return result;
#elif defined(ASM386)
  // *** inline i386 assembler ***
  // damned sign!!
  // x%m will not be normalized by default, but often you depend on it!
  // -1%2 -> -1 , 1%2 -> 1,
  // therefore (-1%2 != 1%2) but one surely expects -1=1 (mod 2) !!!
  register unsigned int result;
  register unsigned int clobbered;
  asm ( \
   "cdq # prepare signed division \n\t"
   "idivl %[mod] # remainder -> edx \n\t" \
   "testl %%edx,%%edx \n\t" \
   "jns 1f # signed? \n\t"
   "addl %[mod],%%edx # normalize to positive value \n\t"
   "1:"
   : "=a" (clobbered), "=&d" (result) : "a" (x), [mod] "g" (m) : "cc");
  return result;
#else
  // *** generic code ***
  // damned sign!!
  // x%m will not be normalized by default, but often you depend on it!
  // -1%2 -> -1 , 1%2 -> 1,
  // therefore (-1%2 != 1%2) but one surely expects -1=1 (mod 2) !!!
  return (x%m<0) ? (x%m)+m : x%m;
#endif
}


#if defined(ASM386) || defined(ASM386_CMOV)
// *** i386 specific assembler code ***
// runs much faster than the generic code and has less contraints :)

#ifdef DEBUG
#warning "i386-assembler code optimizations for mulmod and squaremod enabled"
#endif

/*!
 * @param a an unsigned integer value with @p a
 * @param b an unsigned integer value with @p b
 * @param m an unsigned integer value
 * @result @p a * @p b mod @p m
 * @remark
 *         - This function is often necessary to avoid
 *           integer overflows in the product.
 *         - i386 architecture specific
 *         - for generic usage you should assure @p a < @p m and  @p b < @p m
 */
inline unsigned int mulmod(const unsigned int a, const unsigned int b, const unsigned int m)
{
  // returns (a*b) mod m
  unsigned int x;
  register unsigned int clobbered;
  asm ( \
   "mull %[faktor2] # multiply the factors \n\t" \
   "divl %[mod] # remainder -> edx \n\t"
   : "=a" (clobbered), "=&d" (x) : "a" (a), [faktor2] "g" (b), [mod] "g" (m) : "cc");

  // IMPORTANT REMARK
  // If you request a register for an input operand and this register gets modified,
  // you have to tell GCC that this register is invalid for this operand after leaving the
  // assembler code...
  // You cannot put the register on the clobberlist (because GCC forbids operands to use
  // clobbered registers), 
  // instead you have to declare a dummy output operand that uses the same register...
  // Be careful: if all output operands are dummies, you have to declare the assembler code
  // as volatile!

  return x;
}

/*!
 * @param a an unsigned integer value with @p a < @p m
 * @param m an unsigned integer value
 * @result @p a * @p a mod @p m
 * @remark
 *	   - This function is often necessary to avoid
 *           integer overflows in the product, it should be slightly faster than
 *           the multiplication function.
 *         - i386 architecture specific
 *         - for generic usage you should assure @p a < @p m
 */
inline unsigned int squaremod(const unsigned int a, const unsigned int m)
{
  // returns (a^2) mod m
  unsigned int x;
  register unsigned int clobbered;
  asm ( \
   "mull %%eax # square the factor \n\t" \
   "divl %[mod] # remainder -> edx \n\t"
   : "=a" (clobbered), "=&d" (x) : "a" (a), [mod] "g" (m) : "cc");
  return x;
}

#else
// *** generic variants for mulmod and squaremod ***

/*!
 * @param a an unsigned integer value with @p a < @p m
 * @param b an unsigned integer value with @p b < @p m
 * @param m an unsigned integer value
 * @result @p a * @p b mod @p m
 * @remark This function is often necessary to avoid
 *         integer overflows in the product.
 */
inline unsigned int mulmod(register unsigned int a, register unsigned int b, const unsigned int m)
{
  // assumes a<m, b<m !!
  // hint: multiplication runs faster if b<=a
  // returns (a*b) mod m
  //a%=m; b%=m; // (only necessary if a>=m or b>=m)
  register unsigned int x = (b&1) ? a : 0;
  while (b>>=1)
   {
    a<<=1; if (a>=m) a-=m;
    if (b&1) { x+=a; if (x>=m) x-=m; }
   }
  return x;
}

/*!
 * @param a an unsigned integer value with @p a < @p m
 * @param m an unsigned integer value
 * @result @p a * @p a mod @p m
 * @remark This function is often necessary to avoid
 *         integer overflows in the product, it should be slightly faster than
 *         the multiplication function.
 */
inline unsigned int squaremod(register unsigned int a, const unsigned int m)
{
  // assumes a<m !!
  // returns (a^2) mod m
  //a%=m; // only necessary if a>=m
  register unsigned int b=a;
  register unsigned int x = (b&1) ? a : 0;
  while (b>>=1)
   {
    a<<=1; if (a>=m) a-=m;
    if (b&1) { x+=a; if (x>=m) x-=m; }
   }
  return x;
}

#endif



#if defined(ASM386_CMOV)

#ifdef DEBUG
 #warning "using cmov-operation for powmod(unsigned int, unsigned int, const unsigned int)"
#endif

/*!
 * modular exponentiation
 * @param a base, an unsigned integer value
 * @param pow exponent, an unsigned integer value
 * @param m unsigned integer value
 * @result @p a ^ @p pow mod @p m
 */
inline unsigned int powmod(register unsigned int a, register unsigned int pow, const unsigned int m)
{
  // returns (a^pow) mod m
  // assumes a<m
  register unsigned int x, temp;
  asm ( \
   "movl $1,%[X]       \n\t" \
   "shrl $1,%[pow]     \n\t" \
   "cmovc %%eax,%[X]   \n\t" \
   "jz 1f              \n\t" \
   "0: mull %%eax      \n\t" \
   "divl %[M]          \n\t" \
   "shrl $1,%[pow]     \n\t" \
   "movl %%edx,%%eax   \n\t" \
   "ja 0b  # CF=0 and ZF=0 \n\t" \
   "movl %%edx,%[tmp]  \n\t" \
   "mull %[X]          \n\t" \
   "divl %[M]          \n\t" \
   "movl %%edx,%[X]    \n\t" \
   "movl %[tmp],%%eax   \n\t" \
   "test %[pow],%[pow] \n\t" \
   "jnz 0b             \n\t" \
   "1:  " \
   : [X] "=&r" (x), "+a" (a), [pow] "+r" (pow), [tmp] "=&rm" (temp) : [M] "rm" (m) : "cc", "edx");
  return x;
}

#else
// *** generic variant of powmod ***

/*!
 * modular exponentiation
 * @param a base, an unsigned integer value
 * @param pow exponent, an unsigned integer value
 * @param m unsigned integer value
 * @result @p a ^ @p pow mod @p m
 */
inline unsigned int powmod(register unsigned int a, register unsigned int pow, const unsigned int m)
{
  // assumes a<m
  // returns (a^pow) mod m
  unsigned int x = (pow&1) ? a : 1;
  while (pow>>=1)
   {
     a=squaremod(a,m);
     if (pow&1) x=mulmod(x,a,m);
   }
  return x;
}
#endif


/*!
 * @param p unsigned integer to test for primality
 * @param base base for the strong probable prime test
 * @returns whether @p p is a strong probable prime regarding to @p base .
 */
bool strong_probab_prime(const unsigned int p, const unsigned int base);


/*!
 * @param p unsigned integer to check for primality
 * @result whether @p p is probably prime to the bases 2, 7 and 61.
 * 
 * @remark actually the result returns, whether @p p is prime,
 *         iff 61 < @p p < 4.759.123.141 (and no integer overflow occurs),
 *         as someone has checked these conditions beforehand.
 */
bool probab_prime(const unsigned int p);


/*!
 * @param p unsigned integer to check for primality
 * @result whether @p p is a prime number
 * 
 * @remark This function is slow for large numbers, but safe for all numbers, since
 *         it is based on trial division.
 */
bool is_prime(register const int p);


/*!
 * @result greatest common divisor of @p a and @p b
 */
inline unsigned int gcd(register unsigned int a, register unsigned int b)
{
  // returns greatest common divisor
  while (a&&b) if (a>b) a%=b; else b%=a;
  return a ? a : b;
}

/*!
 * @param a must be an odd unsigned integer
 * @param b must be an odd unsigned integer
 * @result greatest common divisor of @p a and @p b
 * @remark 
 *	   - This function should be somewhat faster than the normal gcd computation
 *           since divisions are replaced by shift operations.
 *         - If the algorithm is feeded with even values, it returns
 *           the odd part of their gcd.
 */
inline unsigned int oddgcd(register unsigned int a, register unsigned int b)
{
#ifdef ASM386_CMOV
 // avoid branches since cmovs are faster on random data and enable parallel execution

#ifdef DEBUG
 #warning "using cmov-operation for oddgcd(unsigned int, unsigned int)"
#endif

  asm ( \
   "1: movl %%eax,%%ecx \n\t" \
   "shrl $1, %%eax \n\t" \
   "cmovcl %%ecx,%%eax \n\t" \
   "movl %%ebx,%%edx \n\t" \
   "shrl $1,%%ebx \n\t" \
   "cmovcl %%edx,%%ebx \n\t" \
   "movl %%eax,%%ecx \n\t" \
   "subl %%ebx,%%eax \n\t" \
   "cmovbe %%ecx,%%eax \n\t" \
   "movl %%ebx,%%edx \n\t" \
   "subl %%eax,%%ebx \n\t" \
   "cmovbe %%edx,%%ebx \n\t" \
   "jnz 1b \n"
   : "=a" (a), "=b" (b) : "a" (a), "b" (b) : "ecx", "edx", "cc");
  return a;
#else
  // *** generic version ***
  // returns greatest (odd) common divisor
  if (a==0) return b;
  if (b==0) return a;
  do
   {
     //cout << a << "," << b << endl;
     while (1&a^1) a>>=1;
     while (1&b^1) b>>=1;
     if (a>b) a-=b;
     if (b>a) b-=a;
   } while (a!=b);
  //cout << a << endl;
  return a;
#endif
}


/*!
 * @param a an unsigned integer
 * @param b an unsigned integer
 * @result whether @p a and @p b are coprime (that is gcd(a,b)==1)
 */
inline bool coprime(register unsigned int a, register unsigned int b)
{
  return ((a|b)&1) ? oddgcd(a,b)==1 : false;
}


/*!
 * @param a an integer value
 * @param b an odd unsigned integer value
 * @result Jacobi symbol (a/b) (which is undefined for even @p b)
 */
signed int jacobi (int a, unsigned int b);


/*!
 * @param a an integer value
 * @param p an odd prime number (as unsigned integer)
 * @result Legendre symbol (a/p), which is
 *          - 1, if @p a is a quadratic residue modulo @p p
 *          - -1, if @p a is not a quadratic residue modulo @p p
 *          - 0, if @p a and @p p are not coprime
 * @remark 
 *         - The result is only defined, if @p p is an odd prime number. You cannot rely on
 *           on any specific behaviour on this function, if this condition is not met!
 *         - For valid values the result is identical to that of the jacobi function.
 */
signed int legendre (signed int a, const unsigned int p);


/*!
 * @param x an unsigned integer value
 * @param m an unsigned integer value
 * @result the modular inverse of @p x mod @p m
 * @remark The result is only defined, if @p x and @p m are coprime.
 */
unsigned int invmod(const unsigned int x, const unsigned int m);


} // namespace numtheory

#endif /* MODULO_H_INCLUDED */
