/* Faktorisierungsprogramm
   written 1997-2000, 2002-2004 by Thorsten Reinecke
   version 2.98: 2004-08 by Thorsten Reinecke
   
   Fr den mathematischen Hintergrund siehe:
   Riesel, Hans: "Prime Numbers and computer methods for factorization", 2nd ed.,
   Boston; Basel; Stuttgart: Birkhuser, 1994
    
   Das Programm setzt einige der dort beschriebenen Algorithmen (in freier
   Variation) um.
   Zunchst findet eine Probedivision fr kleine Primzahlen statt.
   Anschlieend wird mit der Pollard-rho-Methode nach weiteren kleineren
   Faktoren gesucht.
   Falls um eins verminderte Faktoren ausschlielich in kleinere Teiler
   zerfallen, so werden diese mittels der Pollard-Phi-Methode entdeckt.
   
   Wenn die bis dahin verbleibende Zahl nachweislich zusammengesetzt
   (also nicht pseudoprim) ist, wird eine Faktorisierung durch eine
   Variation des multipolynomialen Quadratischen Siebs versucht.
   
   (vgl. Robert D. Silverman: The Multiple Polynomial Quadratic Sieve,
    Mathematics Of Computation, Vol.48, January 1987, p. 329-339)

   Die dort beschriebene Large Prime Variante verwende ich ebenfalls.
   Zustzlich benutze ich aber Large Primes als dynamische Faktoren zur
   dynamischen Erweiterung der (usrprnglich statischen) Faktorbasis, d.h.
   ich siebe zustzlich mit ihnen. Damit das trotz ihrer Gre effizient
   geschehen kann, benutze ich Priority Queues (z.B. Fibonacci-Heaps).
   Diese Methode ist auf eigenem Mist gewachsen; in der von mir studierten
   Literatur konnte ich keinen Hinweis auf hnliche Methoden finden, die
   die Faktorbasis dynamisch erweitern. -- Zustzlich zu diesem Verfahren
   habe ich mich durch ein als "Double Large Prime"-Variante bekanntes
   Verfahren inspirieren lassen. Dieses ist teilweise implementiert und an
   meine dynamische Erweiterung angepat.
   Eine Zyklensuche zur Konstruktion von vollstndig in der Faktorbasis
   aufgehenden Relationen aus Double Large Primes (LP1,LP2 - LP2,LP3 - ...
   ... - LPk,LP1) ist nunmehr ebenfalls implementiert.
   Weitergehende Verbesserungen sind allerdings noch nicht implementiert:
   Eine Absplittung dynamischer Faktoren aus "triple large primes"
   unter Zuhilfenahme von "double large primes"-Zyklen gerade bei der
   Faktorisierung sehr groer Zahlen drfte z.B. ebenfalls sehr ntzlich
   sein, da sie die dynamische Faktorbasis vergrert...

   Das durch die gesiebten Relationen entstehende lineare Gleichungssystem
   wird dabei simultan mit den eingefgten Relationen gelst.
   Um Speicherplatz zu sparen, die Korrektheit bertragener Relationen beim
   verteilten Sieben einfacher berprfen zu knnen sowie triviale Lsungen
   schneller zu entdecken, habe ich die Relationen normiert:
   Jede Relation hat bei mir nun die Form 
    "enthaltene Faktorbasiselemente mit ungeradem Exponenten" * x^2 = 1 (mod n),
   wobei x den bereits ermittelbaren Quadratwurzelanteil darstellt.
   Mehrfachpotenzen der enthaltenen Faktorbasiselemente werden bei
   (ganzzahliger) Halbierung des Exponenten zu x hinzumultipliziert. -- Wenn
   schlielich eine Relation der Form x^2=1 entstanden ist, so kann ein
   (hoffentlich nichttrivialer) Faktor ber gcd(x-1,n) gefunden werden.

   Es ist auch eine Implementierung elliptischer Kurven zur Ermittlung von
   Faktoren "gemigter Gre" enthalten. Dies ist das einzige Verfahren,
   mit dem sich momentan auch bei sehr groen Zahlen nichttriviale Faktoren
   entdecken lassen (da seine Laufzeit von der Gre des entdeckten Faktors
   abhngig ist).

   In seiner jetzigen Fassung sollte man mit dem Programm
   Zahlen bis um 100 Dezimalstellen in ertrglicher Zeit faktorisieren knnen.
  
   [ 
     Nachtrag Juni 2003:

     Bei greren Zahlen lassen sich mit den elliptischen Kurven immerhin
     nichttriviale Faktoren absplitten, z.B. als Dauerproze im Hintergrund:
     P190169 (480 Dezimalstellen); entdeckter Faktor:
     9820062895352727932801465820946503529 (37) [ecm,sigma=272319096]

     Fr die Version 2.90 habe ich einige Testlufe mit folgender
     Konfiguration gestartet: 1x Athlon 1.2 Ghz + 1x Athlon 1.75 Ghz
      + 1x Pentium III 1.0 Ghz + Pentium MMX 166 Mhz.
     Einige dieser Rechner waren zeitweise ber WAN (DSL-Anschlu) angebunden;
     elliptische Kurven wurden fr den Testlauf deaktiviert.
     10^93+3 wurde in knapp 20 Stunden faktorisiert.
     10^95+3 bentigte rund 3 Tage.

     Nachtrag September 2003:
     Im kleinen Netzwerk bentigte die Version 2.92 fr die Faktorisierung einer
     107-stelligen Lucaszahl mit dem Siebverfahren ca. 5 Wochen Rechenzeit.
     (Die Faktorisierung wurde dabei als reiner Hintergrundprozess betrieben.)
   ]
   
   Es fehlt allerdings noch eine ganze Reihe an Optimierungen und Algorithmen.
   Fr ein effizienteres Sieben wre vielleicht noch das "Number-Field-Sieve"
   sinnvoll. (Siebverfahren faktorisieren grundstzlich in Abhngigkeit von
   der Gre der zu faktorisierenden Zahl; sie sind besonders effizient,
   wenn die Zahl lediglich zwei Primteiler hat und diese von der Stelligkeit
   her nicht allzuweit voneinander entfernt liegen.)

   Ob das simultane Lsen des linearen Gleichungssystems sinnvoll ist, oder
   ob man vielleicht effizientere Methoden (z.B. Wiedemann-Verfahren) nach
   Abschlu der Siebphase anwendet, hngt im wesentlichen vom Speicherbedarf
   des Gleichungssystems ab. Beim verteilten Sieben hat das simultane Lsen
   des Gleichungssystems zumindest den Vorteil, parallel zur Siebphase abzu-
   laufen. 
*/

/*! @file
 * @brief
 * main implementation file of "Faktorisierungsprogramm"
 */


#ifdef AUTODETECTION
 // autodetection for gcc-compiler
 #warning "autodetection enabled..."
 #ifdef __GNUG__
  #if __GNUC__ > 2
  #else
    #error "Please compile again with a newer version of GCC (version >= 3.2.x, better >=3.4.x)"
  #endif

  #if #cpu (i386)
   #ifndef ASM386
    #warning "activating i386-inline assembler code"
    #define ASM386
    #ifdef ASMATHLON
     #warning "using athlon-inline assembler code"
     #ifndef ASM386_CMOV
     #warning "activating ASM386-CMOV usage, because ATHLON supports it :)"
     #define ASM386_CMOV
     #endif
    #endif
   #endif
  #else
   #ifdef ASM386
    #warning "deactivating i386-inline assembler code"
    #undef ASM386
    #undef ASM386_CMOV
    #undef ASMATHLON
   #endif
  #endif
  #if !defined (unix) && defined(USE_NETWORK)
    #error "network support only for Unix-systems" 
  #endif
 #endif
#endif /* autodetection */

using namespace std;


#ifndef USE_NETWORK
 #ifdef IS_SERVER
  #error "Illegal configuration: SERVER and no USE_NETWORK"
 #endif
#else
 #ifndef IS_STANDALONE
  #ifdef IS_SERVER
   #include <map>
  #endif
  #include "unix_buffer.H"
 #endif
#endif

#ifndef IS_SERVER
 #ifndef IS_CLIENT
  #undef IS_STANDALONE
  #define IS_STANDALONE
 #endif
#endif

#ifdef IS_SERVER
 #ifdef IS_CLIENT
  #error "Illegal Configuration SERVER-CLIENT"
 #else
  #ifdef IS_STANDALONE
   #error "Illegal Configuration SERVER-STANDALONE"
  #endif
 #endif
#endif

#ifdef IS_CLIENT
 #ifdef IS_STANDALONE
  #error "Illegal Configuration CLIENT-STANDALONE"
 #endif
#endif

#if defined(USE_FIBHEAP) && defined(USE_FAKEHEAP)
 #error "defines USE_FIBHEAP and USE_FAKEHEAP exclude each other!"
#endif

#ifdef REACT_ON_SIGUSR
 // a little class that catches signal SIGUSR1 and SIGUSR2
 // and stores these flags...
 // can be useful to abort some long-running algorithms
 // without aborting the program or altering configuration parameters.
 #include "usr_signals.cc"
 Cusr_signal_proxy USRSignalHandler;
#endif


// propagate IS_STANDALONE-condition from compiletime to runtime,
// so that certain modules can be compiled independently of it.
const bool compiled_IS_STANDALONE()
{
  #ifdef IS_STANDALONE
   return true;
  #else
   return false;
  #endif
}


extern "C"
{
 #include <unistd.h>
}

#include <cstdio>
#include <cstdlib>
#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <cmath>
#include <gmp.h>
#include <list>
#include <set>
#include "utils.H"
#include <ctime>

#include "qsieve.H"


struct myintcompare
{
  inline const bool operator() (const int i1, const int i2) const
   {
     return i1 < i2;
   }
};
typedef set<int, myintcompare> intset;


class TFoundFactor
{
 private:
  mpz_t x;
  unsigned int e;
  string comment;
 public:
  explicit TFoundFactor(const mpz_t f, const unsigned int exp = 1, const string &xcomment="")
   : e(exp), comment(xcomment) { mpz_init_set(x,f); }
  explicit TFoundFactor(const long unsigned int f, const unsigned int exp = 1)
   : e(exp), comment() { mpz_init_set_ui(x,f); }
  TFoundFactor(const TFoundFactor &FF)
   : e(FF.e), comment(FF.comment) { mpz_init_set(x,FF.x); }
  explicit TFoundFactor(std::istream &is)
   {
     mpz_init(x);
     is.get();
     if (is.eof()) { is.setstate(std::ios::failbit); return; }
     is.unget();
     string s;
     is >> x;
     is >> s; if (s!="e") { MARK; is.setstate(std::ios::failbit); return; }
     is >> e;
     is >> s; if (s!="c") { MARK; is.setstate(std::ios::failbit); return; }
     char buf[4096];
     is.getline(buf,sizeof(buf),'\n'); comment=buf;
   }
  virtual ~TFoundFactor() { mpz_clear(x); }
  virtual bool operator< (const TFoundFactor &nf) const { return (mpz_cmp(x,nf.x)<0); }
  virtual void output(std::ostream &ostr) const;
  TFoundFactor& operator= (const TFoundFactor &rhs)
   {
     mpz_set(x,rhs.x); e=rhs.e; comment=rhs.comment;
     return *this;
   }
  virtual void Save(std::ostream &os) const
   {
     os << x << " e " << e << " c " << comment << endl;
   }
};



inline std::ostream& operator << (std::ostream& ostr, const TFoundFactor FF)
{
  FF.output(ostr);
  return ostr;
}

// quick and dirty:
class TFoundFactors : public set<TFoundFactor>
{
public:
 string regarding; // argument (given number, can be symbolic)

 void Load(std::istream &is)
  {
    clear();
    while (is)
     {
MARK;
       TFoundFactor FF(is);
       if (is.good()) insert(FF);
     }
  }
 void Save(std::ostream &os) const
  {
    for (TFoundFactors::const_iterator p=begin(); p!=end(); ++p) p->Save(os);
  }

 void AutoSave() const;
 void AutoLoad();
};

TFoundFactors FoundFactors;


#include <vector>
#include <stack>
#include <algorithm>

#ifdef USE_NETWORK
 extern "C"
 {
  #include <sys/utsname.h>
  #include <pthread.h>
 }
 #include <csignal>
 #include "mutex.H"
 #ifdef IS_CLIENT
  extern "C"
  {
   #include <sys/wait.h>
   #include <unistd.h>
  }
 #endif
#endif

#ifdef NOTIFY_PARENT
#if defined(IS_SERVER) && defined(USE_NETWORK) 
 #include <csignal>
#else
 #warning "undefining NOTIFY_PARENT, since this is not a networked server"
 #undef NOTIFY_PARENT
#endif
#endif


// output to screen/terminal can optionally be handled by a ncurses based
// wrapper, but on default all windows will be merged into cout.
#ifdef USE_NCURSES
 #include <my_ncurses.H>
#endif

ostream cout_titel(cout.rdbuf());
ostream cout_network(cout.rdbuf());
ostream cout_status(cout.rdbuf()); 


const string RecoveryFile         = "recovery.dat";
const string FoundFactorsFile     = "FoundFactors.dat";
const string FactorizationsFile   = "factorizations.txt";
const string StaticRelationsFile  = "static_relations.dat";
const string SpecialRelationsFile = "special_relations.dat";
      string DynamicRelationsFile = "dynamic_relations.dat"; // default, may be overwritten by entry in configfile
      string ConfigFile           = "qsieve.cfg"; // default, may be overwritten by getenv("QSIEVE_CFG")


// Streams fr die Speicherung der Relationen und sonstiger Daten
// zur Verfgung stellen
// *************************************************************************
// **** wichtig:
// **** Wenn ostream und istream durch denselben filebuf fest miteinander
// **** gekoppelt sind, darf/kann man sich *nicht* darauf verlassen, dass
// **** tellp() und tellg() voneinander unabhngige Funktionen sind!
// **** *Jede* Schreiboperation kann dann *auch* den Wert von tellg()
// **** verndern und *jede* Leseoperation den von tellp()!
// ************************************************************************* 

// Fr die Relationen der dynamischen Faktorbasis
// streams fr das Datenfile zur Verfgung stellen:
filebuf DynamicRelations_buffer;
ostream DynamicRelations_to_file (&DynamicRelations_buffer);
istream DynamicRelations_from_file (&DynamicRelations_buffer);

// Fr Recovery
filebuf Recovery_buffer;
ostream Recovery_to_file (&Recovery_buffer);
istream Recovery_from_file (&Recovery_buffer);

#ifdef IS_CLIENT
string communication_name;
ofstream communication_stream;
#endif

#ifndef IS_CLIENT
// Fr das Speichern der gemachten Faktorisierungen
std::ofstream Factorization_to_file;
/* Bemerkung:
   Auf diesem Stream werden alle mit diesem Programm gemachten Faktorisierungen
   protokolliert. Die zu faktorisierenden Zahlen sowie alle gefundenen Faktoren
   werden angehngt.
   Das zugehrige File sollte im Bedarfsfall gelscht werden, wenn die Daten
   nicht mehr bentigt werden.
*/
#endif

#include "mpqsPolynom.H"


#ifndef IS_SERVER
#ifdef USE_FIBHEAP
 #include "fibheap.H"
 FibHeap <TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
 #include <queue>
#elif defined(USE_FAKEHEAP)
 #include "fakeheap.H"
 FakeHeap <TSieve_Delta> Sieve_Delta_Heap;
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_HeapOfSquares;    
#else
 #include <queue>
 priority_queue<TSieve_Delta> Sieve_Delta_Heap, Sieve_Delta_HeapOfSquares;
#endif
#endif

mpz_t n, /* zu faktorisierende Zahl */
      kN, /* das mit einem gnstigen Vorfaktor multiplizierte n
	     (nachdem n um einfach zu findende Faktoren reduziert wurde) */
      r; /* Quadratwurzel von n (oder bei optimierter Fassung k*n) */

int Vorfaktor_von_kN; // der noch zu ermittelnde Vorfaktor fr kN (kN=Vorfaktor_von_kN*n)


int PrimeNumbers[StaticFactorbase::MaxSize]; /* fr eine effiziente Probedivision
				       alle Primzahlen der Faktorbasis hier
				       akkumulieren */


int SQRT_kN_of_PrimeNumbers[StaticFactorbase::MaxSize];
int Delta_of_PrimeNumbers[StaticFactorbase::MaxSize][2]; /* fr den effizienten Zugriff
						    werden die pro Siebintervall berechneten
						    Werte hier abgelegt */
 
int NumberOf_more_PrimePowers = 0; // tatschliche Anzahl (wird im Programm ermittelt)
int FB_maxQuadrate = 0; /* tatschliche Anzahl der Quadrate,
			   mit denen normal gesiebt wird (wird im Programm ermittelt) */
int PrimePowers[StaticFactorbase::max_additional_Powers];
int SQRT_kN_of_PrimePowers[StaticFactorbase::max_additional_Powers];
int Delta_of_PrimePowers[StaticFactorbase::max_additional_Powers][2];
int SQRT_kN_of_PrimeSquares[StaticFactorbase::MaxSize];

#include "StaticRelations.H"


/* Zur Erluterung der Begrifflichkeit von Dynamic-Factors/Special-Factors:
   Ich verwende momentan eine modifizierte Multi-Large-Prime-Variante:
   Dynamic-Factors entsprechen der Large-Prime-Variante. Aber ich siebe mit jedem neu entdeckten
   dynamischen Faktor zustzlich...
   Special-Factors sind zusammengesetzte Restfaktoren (wie sie in der Double-Large-Prime-Variante
   auftreten). Es wird einerseits versucht, Special-Factors durch neu aufgenommene
   Dynamic-Factors zu splitten, um daraus weitere Dynamic-Factors zu gewinnen.
   Andererseits wird eine Zyklensuche auf diesen durchgefhrt, um weitere
   Relationen zu gewinnen.
   In einer spteren Implementierung knnte man die Zyklensuche noch
   auf Hyperkantenzyklen hherer Large-Primes erweitern.
*/

typedef set<TDynamicFactorRelation,TDynamicFactorRelation> TDynamicFactorRelations;

#ifdef IS_SERVER
class CServerDynamicFactorRelations : public TDynamicFactorRelations
{
 private:
  mutable CMutex monitor_mutex;
  vector<int> DynamicFactorList;
  int size_active_SLP, size_passive_SLP;
 public:
  inline void insert(const TDynamicFactorRelation& x)
   {
     if (TDynamicFactorRelations::insert(x).second)
      {
        if (!x.sieveable())
         {
           // x is a Single Large Prime, but a passive one: not sieveable
           ++size_passive_SLP;
           return;
         }
        monitor_mutex.lock();
        ++size_active_SLP;
        DynamicFactorList.push_back(x.factor);
        monitor_mutex.unlock();
      }
     else
      {
        cerr << "CServerDynamicFactorRelations::insert duplicate (rejected)!" << endl;
        exit(1);
      }
   }
  inline const int monitoredSize() // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int s=DynamicFactorList.size();
     monitor_mutex.unlock();
     return s;
   }
  inline const int size_active() const { return size_active_SLP; }
  inline const int size_passive() const { return size_passive_SLP; }
  inline const int operator[] (const int i) // const /* const attribute could be dangerous due to multithreading!? */
   {
     monitor_mutex.lock();
     const int x=DynamicFactorList[i];
     monitor_mutex.unlock();
     return x;
   }
};
CServerDynamicFactorRelations DynamicFactorRelations;
#else
TDynamicFactorRelations DynamicFactorRelations;
#endif

double Factor_Threshold = 3.0; /* Exponent, der grere Faktoren ber die
				  Faktorbasis hinaus erlaubt.
				  Sinnvolle Werte: 1.0 bis 3.0
				  1.0: keine zustzlichen Faktoren auerhalb der Faktorbasis 
				  bis 2.0: Primfaktoren auerhalb der Faktorbasis bis
				  zum Quadrat des grten Faktors der FB
				  ber 2.0: Faktoren bis zu grter Faktor der Faktorbasis
				  hoch Factor_Threshold.
			       */


#include "SpecialRelations.H"


bool collecting_phase_finished = false; /* switch, der die "aktive" und "passive" Phase des Siebens
					  bestimmt:
					  Wenn gengend Special-Relations und Dynamic-Relations gesammelt
					  sind, um ausreichend viele Static-Relations zu erhalten, wird
					  das Aufsammeln weiterer Special- und Dynamic-Relations durch
					  das Setzen von collecting_phase_finished=true unterbunden. */

int biggest_Prime_in_Factorbase;


#include "Sieving.H"
#ifdef IS_SERVER
 // we have to define some variables, because
 // Sieving.o is not linked to the binary for servers...
 int LogicalSieveSize = 1000000; // das zu jedem Polynom gehrige Siebintervall ist [-LogicalSieveSize,LogicalSieveSize]
#endif


// einige Parameter fr easy-Faktorisierung (defaults)
bool SkipFermat    = false; // Fermatmethode versuchen
bool SkipPhi       = false; // Pollard-Phi-Methode versuchen
bool SkipFibonacci = false; // Fibonacci-Methode versuchen
bool SkipEasyECM   = false; // elliptic curve method im easy-mode versuchen
bool SkipECM       = false; // elliptic curve method versuchen
int rho_Phase      =   150000; // Grundinitialisierung fr Rho-Phase
int phi_Phase1     =   500000; // fr phi-Methode Phase 1
double phi_Phase2  = 10000000.0; // fr phi-Methode Phase 2 (=isolierter Faktor)
int elcu_Phase1    =    20000; // fr elliptische Kurven Phase 1
double elcu_Phase2 =  2000000.0; // fr elliptische Kurven Phase 2
int elcu_Kurven    =      100; // Anzahl der zu probierenden elliptischen Kurven
int ecm_curves_processed = 0; // Anzahl der (bereits) durchlaufenden Kurven


CmpqsPolynom Polynom; // Objekt zur Verwaltung der Polynomberechnung fr multipolynomiales Sieb
#include "mpqsStatistics.cc" // statistical stuff about found relations, client connections, etc.


inline bool is_dynamic_factor(const int number)
{
  // returns, whether this number is already registered as a dynamic factor;
  // (dynamic factor -> prime number outside the static factorbase, also known as Single Large Prime)
  TDynamicFactorRelation relation;
  relation.factor = number;
  const TDynamicFactorRelations::iterator pos = DynamicFactorRelations.find(relation); // dynamischen Faktor suchen
  return (pos != DynamicFactorRelations.end());
}


#include "modulo.H" // modulo operations fr unsigned int
using namespace numtheory;


const int probab_prime_checks = 25; // #rounds for mpz_probab_prime

string XML_StatusFile;


void TFoundFactor::output(std::ostream &ostr) const
{
  ostr << x;
  if (e!=1) ostr << "^" << e;
  if (comment!="") ostr << " " << comment;
  else
   {
     if (!mpz_probab_prime_p(x,probab_prime_checks)) ostr << "[composite]";
   }
}

void TFoundFactors::AutoSave() const
{
  std::ofstream ofstr(FoundFactorsFile.c_str());
  ofstr << regarding << endl;
  Save(ofstr);
}

void TFoundFactors::AutoLoad()
{
  cout << "Autoload FoundFactors" << endl;
  std::ifstream ifstr(FoundFactorsFile.c_str());
  char buf [32768];
  ifstr.getline(buf,sizeof(buf),'\n'); regarding=buf;
  Load(ifstr);
}


#ifndef IS_CLIENT
 #include "ConfigFile.cc"
#endif


const char* const _MAL(void)
{
  // liefert Leerstring bei erstem Aufruf, " * " sonst.
  // Aufgabe: Verhinderung aufwendiger Tests, um den Faktor 1 als
  // ersten bzw. letzten Faktor unterdrcken zu knnen.

#ifdef NOTIFY_PARENT
  // send SIGUSR2 to the parent to notify that a new
  // factor has found
  kill(getppid(),SIGUSR2); // send SIGUSR2 to the parent process
#endif

  static bool erster_Aufruf = true;
  if (erster_Aufruf)
   {
     erster_Aufruf=false;
     return "";
   }
  else return " * ";
}

std::string MAL(const mpz_t factor, const unsigned int exponent=1)
{
  FoundFactors.insert(TFoundFactor(factor,exponent));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str();
}

std::string MAL(const mpz_t factor, const unsigned int exponent, const std::ostringstream &comment)
{
  FoundFactors.insert(TFoundFactor(factor,exponent,comment.str()));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str()+comment.str();
}


std::string MAL(const mpz_t factor, const std::ostringstream &comment)
{
  FoundFactors.insert(TFoundFactor(factor,1,comment.str()));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  return helper.str()+comment.str();
}

std::string MAL(const long unsigned int factor, const unsigned int exponent=1)
{
  FoundFactors.insert(TFoundFactor(factor,exponent));
  FoundFactors.AutoSave();
  ostringstream helper;
  helper << _MAL() << factor;
  if (exponent>1) helper << "^" << exponent;
  return helper.str();
}


#include "polynomial.H" /* for fast polynomial arithmetic & fft-continuation */
#include "fft_param.cc"  /* discrete fast fourier transform */
#include "elliptic_curve.H" /* invariant part of elliptic curve method (ecm) */
#include "elliptic_curve-variant.cc" /* variant part of elliptic curve method (ecm) */

#include "ExitManager.cc" /* controlled termination of program */
#include "StaticRelations.cc"

#ifndef IS_CLIENT
 // this stuff may only be needed by server or stand-alone version
 #include "easy_factor.H"
#endif /* of #ifndef IS_CLIENT */


#include "sqrt_modulo.cc"

inline int SQRT_kN_mod_PrimeNumber(const unsigned int Primzahl)
{
  return sqrtmod(mpz_remainder_ui(kN,Primzahl),Primzahl);
}


int check_SQRT_kN_mod_PrimeNumber(const int Primzahl)
{
  // kann zu Debugging-Zwecken aufgerufen werden, um zu testen,
  // inwieweit die "SQRT_kN_mod_PrimeNumber"-Funktion korrekt arbeitet
  int w1 = SQRT_kN_mod_PrimeNumber(Primzahl);
  if (w1>Primzahl/2) w1=Primzahl-w1;

  mpz_t x;
  mpz_init_set_ui(x,w1); mpz_mul(x,x,x); mpz_sub(x,x,kN);
  if (mpz_mod_ui(x,x,Primzahl)!=0)
    {
      MARK;
      cerr << "wrong result! " << w1 << " mod " << Primzahl << endl;
      exit(1);
    }
  mpz_clear(x);
  return w1;
}


void compute_StaticFactorbase()
{
#if defined(ASMATHLON) || defined(ASM_P3_SSE)
  // ATHLON specific sanity checks
  if ( LogicalSieveSize>(1<<23) || (StaticFactorbase::Size()&3) )
   {
     MARK;
     cerr << "Sorry! Processor specific constraints not met." << endl;
     cerr << "For using AMD-3DNow! or SSE optimized version we need:" << endl;
     cerr << "Constraint 1 (SIMD): (Size_StaticFactorbase&3)==0" << endl;
     cerr << "Constraint 2 (float precision): LogicalSieveSize<=(1<<23) [=8388608]" << endl;
     cerr << "Please edit ConfigFile according to these constraints" << endl;
     cerr << "OR use a version without these optimizations." << endl;
     exit(1);
   }
#endif

  cout << "Calculating static factorbase." << endl;
  PrimeNumbers[0]=-1; /* nullte "Primzahl" ist -1 (Bercksichtigung des
		       Vorzeichens bei den Relationen! */
  SieveControl::log_PrimeNumbers_dirty[0]=SieveControl::log_PrimeNumbers[0]=0; // Vorzeichen hat keinen Logarithmuswert

  int Primzahl=2;
  // die Primzahl 2 ist ein Sonderfall. Ich gehe davon aus, dass sie
  // auf jeden Fall in der Faktorbasis vorhanden sein soll.
  PrimeNumbers[1]=2;
  SieveControl::log_PrimeNumbers_dirty[1]=SieveControl::log_PrimeNumbers[1]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)));

  Primzahl=1; // Initialwert 1 nur, wenn in der Schleife +=2 gezhlt wird!
  mpz_t x,y,wuqu;
  mpz_init(x); mpz_init(y); mpz_init(wuqu); // Fr Berechnung der Quadratwurzeln von Primzahlpotenzen
  for (int Primzahl_nr=2; Primzahl_nr<StaticFactorbase::Size(); Primzahl_nr++)
    {
      do // nchsthhere Primzahl der statischen Faktorbasis ermitteln 
	{
	  do Primzahl+=2; while(!is_prime(Primzahl)); /* nchste Primzahl */
	  mpz_set_ui(x,Primzahl);
	}
      while (mpz_legendre(kN,x)<0); /* Primzahl gltig fr Primzahlbasis? */
      
      //printf("%i. Primzahl der Faktorbasis ist %i\r",Primzahl_nr,Primzahl);
      PrimeNumbers[Primzahl_nr]=Primzahl;
      
      // fr effizientes Sieben erforderlich (erspart dauernde Neuberechnung)
      SieveControl::log_PrimeNumbers[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)));
      SieveControl::log_PrimeNumbers_dirty[Primzahl_nr]=static_cast<TSieveElement>(ceil(log_Multiplikator*log(Primzahl)*SieveControl::DirtyFactor(Primzahl_nr)));
      
      // fr sptere Ermittlung der Deltawerte zum Sieben erforderlich
      SQRT_kN_of_PrimeNumbers[Primzahl_nr]=SQRT_kN_mod_PrimeNumber(Primzahl);
      
      // ggf. Prfung, ob Wurzel korrekt ist:
      if ( squaremod(SQRT_kN_of_PrimeNumbers[Primzahl_nr],Primzahl) != mpz_remainder_ui(kN,Primzahl) )
       {
         cerr << "square root not correct!" << endl;
         cerr << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << "^2 <> "
              << mpz_remainder_ui(kN,Primzahl) << " (mod " << Primzahl << ")" << endl;
         exit(1);
       }


      if ( Primzahl < sqrt(std::numeric_limits<int>::max()) && Vorfaktor_von_kN%Primzahl!=0 )
       {
          // *** Wurzeln von kN modulo Primzahl^2 ermitteln ***
          const unsigned int P = Primzahl*Primzahl;
	  const unsigned int Wu = SQRT_kN_of_PrimeNumbers[Primzahl_nr];
          unsigned int iy = invmod(2*Wu,P);
          unsigned int ix = mpz_remainder_ui(kN,P);
          
	  unsigned int wuq=Wu*Wu;
          if (ix>=wuq) ix-=wuq; else ix=ix-wuq+P; // ix-=wuq; berlauf verhindern; wir rechnen modulo P
          ix=mulmod(ix,iy,P)+Wu;
          wuq = (ix<P) ? ix : ix-P; // schnellere Version fr: wuq=ix%P

#if 1 || defined(DEBUG)
          if (mulmod(wuq,wuq,P)!=mpz_remainder_ui(kN,P))
           {
             cerr << "PQ-root incorrect!" << endl;
             cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << endl;
             cerr << "kN= " << kN << endl;
	     cerr << mulmod(Wu,Wu,P) << " != " << mpz_remainder_ui(kN,P) << endl;
             exit(1); 
           }
#endif
          SQRT_kN_of_PrimeSquares[Primzahl_nr]=wuq;
       } else SQRT_kN_of_PrimeSquares[Primzahl_nr]=-1;


      const int Threshold = LogicalSieveSize < std::numeric_limits<int>::max()/8 ? 8*LogicalSieveSize/Primzahl : LogicalSieveSize/Primzahl;
      if (Primzahl<=Threshold && Primzahl>2 && mpz_remainder_ui(kN,Primzahl)!=0)
	// Primzahl darf nicht zu gro sein, ungleich 2 und kein Teiler von kN, um mit Potenzen gesiebt zu werden
	{
	  FB_maxQuadrate=Primzahl_nr; /* nach Abschlu der Schleife ist dies der Wert, bis zu dem
					 Quadrate auf normale Weise gesiebt werden. */

	  // Wurzeln von kN modulo Primzahl^pot ermitteln
	  // Initialwerte:
	  int P_Potenz = Primzahl*Primzahl;
	  int Wu       = SQRT_kN_of_PrimeNumbers[Primzahl_nr];

          double Wichtung = Primzahl_nr<SieveControl::sieben_ab_Primzahlindex() ? 2.0 : 1.0;
           // mit den ersten Primzahlen der Faktorbasis wird (weil sie zu klein sind) gar nicht
           // gesiebt; dafr erhlt die kleinste Potenz dieser Zahl, die hier aufgenommen wird,
           // den vollen Siebwert der Potenz: log(P_Potenz)=Potenz*log(Primzahl).
           // Alle brigen Potenzen entsprechen dem nochmaligen Auftreten eines schon bekannten
           // Faktors und knnen deshalb nur in einfacher Zhlung (=log(Primzahl) bercksichtigt werden.

	  for (int pot=2; ;pot++)
	    {
	      //cout << "berechne Wurzel fr " << Primzahl << "^" << pot << endl;

	      mpz_set_ui(y,2*Wu); mpz_set_ui(x,P_Potenz);
	      if (mpz_invert(y,y,x)==0)
		{ 
		  cerr << "PPot-root for " << Primzahl  << ": inverse doesn't exist!" << endl;
		  break;
		}

              if (NumberOf_more_PrimePowers>=StaticFactorbase::max_additional_Powers)
               { 
                 static char weitermachen = 'n';
                 if (weitermachen!='y')
                  {
	           cerr << "Overflow for " << Primzahl << "^" << pot << endl;
                   cerr << "Es kann nicht optimal mit Primzahlpotenzen gesiebt werden," << endl
                        << "da der hierfr reservierte Speicherplatz nicht ausreicht." << endl;
                   cerr << "Sie sollten \"StaticFactorbase::max_additional_Powers\" vergroessern!" << endl;
                   cerr << "Reserved memory for storing prime powers is too small..." << endl;
                   cerr << "sieving will be less efficient..." << endl;
                   cerr << "continue? (y/n) " << flush;
                   cin >> weitermachen;
                   if (weitermachen=='n') exit(1);
                  }
                 break; // berlauf durch Schleifenabbruch verhindern
               }

	      mpz_mod_ui(x,kN,P_Potenz); mpz_set_ui(wuqu,Wu); mpz_mul_ui(wuqu,wuqu,Wu); mpz_sub(x,x,wuqu);
	      mpz_mul(x,x,y); mpz_add_ui(x,x,Wu); mpz_mod_ui(x,x,P_Potenz);
	      Wu = mpz_get_ui(x); if (Wu>P_Potenz-Wu) Wu=P_Potenz-Wu; // auf die "kleinste" Wurzel normieren
              // Wu ist nun die Quadratwurzel von kN (mod P_Potenz)

	      // ggf. Prfung, ob Wurzel korrekt ist:
	      mpz_set_ui(x,Wu); mpz_mul(x,x,x); mpz_sub(x,kN,x); 
	      if (mpz_mod_ui(x,x,P_Potenz)!=0)
		{
		  cerr << "PPot-root incorrect!" << endl;
		  cerr << "Primzahl=" << PrimeNumbers[Primzahl_nr] << " Nr. " << Primzahl_nr << endl;
		  cerr << "kN= " << kN << endl;
		  cerr << "SQRT(kN) mod p=" << SQRT_kN_of_PrimeNumbers[Primzahl_nr] << endl;
		  cerr << "SQRT(kN) mod p^i=" << SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers] << endl;
		  cerr << "but we got a wrong value: " << x << endl;
		  exit(1);
		}

              // Heuristik: Fr die letzte Potenz die Wichtung etwas erhhen,
              // da hier eine gewisse Chance besteht, dass der Treffer auch Treffer noch hherer Potenzen ist
              if (P_Potenz>Threshold) Wichtung+=4.0/Primzahl;

#if 1
              // Heuristik (im Versuch gemessen/ermittelt):
              // Bei "kleinen" Zahlen geht das Sieben deutlich schneller, wenn man die kleineren
              // Zahlen wegfallen lsst; das Sieb wird dadurch allerdings durchlssiger, aber das strt nicht,
              // da die Relationen nur so reinpurzeln...
              // Bei greren Zahlen werden aber bei genauerem Sieben in der gleichen Zeit mehr DLP erkannt, und
              // das ist dort der wichtige Effekt.
              if ( P_Potenz<250 && mpz_sizeinbase(kN,10)<70 ) Wichtung+=1.0;
              else
#endif
               {
                 // weil das Sieben mit kleinen Zahlen viel Zeit bentigt und ein Siebtreffer wenig
                 // bringt, verlangen wir eine Mindestgre, damit die Potenz beim Sieben
                 // bercksichtigt wird.

	         PrimePowers[NumberOf_more_PrimePowers]=P_Potenz;
	         SQRT_kN_of_PrimePowers[NumberOf_more_PrimePowers]=Wu;

		 log_Primzahl_of_PrimePowers[NumberOf_more_PrimePowers]
                   =static_cast<TSieveElement>(ceil(log_Multiplikator*Wichtung*log(Primzahl)));
                 Wichtung=1.0;
                 ++NumberOf_more_PrimePowers; // dies war eine weitere aufgenommene Primzahlpotenz
               }
	      
	      if (P_Potenz>Threshold) break; // nchste Primzahlpotenz wird zu gro...
	      P_Potenz*=Primzahl;
	    }
	}
    }
  mpz_clear(x); mpz_clear(y); mpz_clear(wuqu);

  biggest_Prime_in_Factorbase=Primzahl;
  cout << "prime numbers in static factorbase have been computed." << endl;
  
  if (Primzahl>=PhysicalSieveSize)
    cout << "Remark: Some members of the static factorbase are bigger than the sieve size!" << endl;
}


#ifndef IS_SERVER
void compute_Deltas_for(const TDynamicFactorRelation &DynRel, const int offset)
{
  TSieve_Delta sieb_delta;
  int d1,d2;
  
  sieb_delta.factor=DynRel.factor;

  const unsigned int inv_A2 = invmod(mpz_remainder_ui(Polynom.get_A2(),DynRel.factor),DynRel.factor);
  const unsigned int PolyB = mpz_remainder_ui(Polynom.get_B(),DynRel.factor);

  d1=DynRel.SQRT_kN-PolyB; if (d1<0) d1+=DynRel.factor;
  d1=mulmod(d1,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d1-offset,DynRel.factor);
  sieb_delta.delta+=offset;
#ifdef USE_FAKEHEAP
  while (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird
   {
     Sieve_Delta_Heap.push(sieb_delta);
     sieb_delta.delta+=sieb_delta.factor;
   }
#else
  if (sieb_delta.delta<=LogicalSieveSize) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
    Sieve_Delta_Heap.push(sieb_delta);
#endif
  
  d2=DynRel.factor-DynRel.SQRT_kN-PolyB; if (d2<0) d2+=DynRel.factor;
  d2=mulmod(d2,inv_A2,DynRel.factor);
  sieb_delta.delta=normalized_signed_mod(d2-offset,DynRel.factor);
  sieb_delta.delta+=offset;
  if ( (sieb_delta.delta<=LogicalSieveSize)  && (d1!=d2) ) // shortcut: nur dann eintragen, wenn die Relation noch im Siebintervall relevant wird 
   {
     Sieve_Delta_Heap.push(sieb_delta);
#ifdef USE_FAKEHEAP
     sieb_delta.delta+=sieb_delta.factor;
     while (sieb_delta.delta<=LogicalSieveSize)
      {
        Sieve_Delta_Heap.push(sieb_delta);
        sieb_delta.delta+=sieb_delta.factor;
      }
#endif
   }
}
#endif




// fr die Speicherung der gefundenen dynamischen Faktoren.
// Bemerkung: Um die dynamischen Faktoren bereits in der Siebphase
//            aufzusammeln, ist das Array global.
//             (Dafr wird die Anzahl der gefundenen dynamischen Faktoren
//              fr die aktuelle Relation dem Konstruktor bergeben.)
const short int MaxHits=64; // bei Bedarf erhhen...
static struct {int Faktor, Exponent;} Hits[MaxHits]; // Dynamische Special-Factoren der Relation aufsammeln
// wichtiger Hinweis: Der Exponent zum Faktor wird nicht beim Sieben, sondern erst bei der Refaktorisierung
// ermittelt. Falls also mit Primzahlpotenzen gesiebt werden sollte, braucht der Exponent nicht eingetragen
// zu werden. Als Faktor sollte in jedem Falle die Primzahl und nicht die Primzahlpotenz eingetragen werden!


#ifndef IS_SERVER

CRelation::CRelation(const signed int Siebpos, short int HitCount /* =0 */)
: CProvideHelperVariables(), relevant_factor(no_factor), Relation_sparse(new CTinyFBsizetypeVector), Relation_dense(NULL)
  /* Konstruktor: neue Relation einfgen.
     "Deltas" mit dem Wert Delta initialisieren,
     "Relationen" mit den Primfaktoren des reduzierten Quadrates fllen,
     deren Exponenten ungerade sind. */
{
  // cout << "initialisiere Gleichungszeile" << endl;
  mpz_init(Delta);
  Polynom.get_values(Siebpos,Delta,y);
  // in y: reduziertes Quadrat (ist nun zu faktorisieren...) 
  mpz_invert(Delta,Delta,n); /* wir verwenden den Kehrwert von Delta,
				das erspart eine zustzliche Variable;
				gleichzeitig ist damit die zu findende
				quadratische Kongruenz auf Delta^2=1 (mod n)
				normiert. */
  
  if (mpz_sgn(y)<0) // Vorzeichenbehandlung 
    {
      Relation_sparse->append(0); 
      /* "nullte Primzahl" ist -1 wegen Bercksichtigung
         des Vorzeichens. Also die 0 fr Position von  -1 einfgen */
      mpz_neg(y,y); /* -1 war Teiler -> positiv machen */
    }
  
  // ersten Faktor der statischen Faktorbasis ebenfalls gesondert behandeln,
  // denn die generierten Polynome haben Probleme, die Primzahl 2 zu sieben...

  if (PrimeNumbers[1]==2)
   {
     // special treatment for 2, since it it very easy:
     const unsigned int z = mpz_scan1(y,0);
     if (z)
      {
        mpz_tdiv_q_2exp(y,y,z);
        mpz_mul_2exp(Delta,Delta,z>>1);
        if (z&1) Relation_sparse->append(1); // ungerade Potenz der Primzahl markieren
      }
   }
  else
   {
     // this is the common case (which works also for 2)  
     unsigned int z=0;
     while (mpz_div_ui(x,y,PrimeNumbers[1])==0) { mpz_set(y,x); z++; }

     if (z&1) Relation_sparse->append(1); // ungerade Potenz der Primzahl markieren
     if (z>1) // Delta mit den "gltigen Wurzeln" multiplizieren
      {
        if (z<4) mpz_mul_ui(Delta,Delta,PrimeNumbers[1]);
        else
         {
	   mpz_set_ui(x,PrimeNumbers[1]); mpz_powm_ui(x,x,z/2,n);
	   mpz_mul(Delta,Delta,x);
         }
      }
   }

  {
    // find factors in static factorbase
    // (beginnend mit der 3. Primzahl der FB (s.o) )
    vector<unsigned int> FB_Primteiler;
    GetStaticFactorHits(FB_Primteiler,Siebpos);

    // Faktoren zusammenmultiplizieren und dann in einem Rutsch abdividieren
    mpz_set_ui(x,1);
    for (unsigned int i=0; i<FB_Primteiler.size(); i++) mpz_mul_ui(x,x,PrimeNumbers[FB_Primteiler[i]]);
    mpz_divexact(y,y,x);

    // new factor reduction using the principle of inclusion/exclusion
    // die Faktor-Potenzen nach dem Prinzip Inklusion/Exklusion abdividieren
    vector<unsigned int> WithOddExponent, TempTeiler;
//    cout << "------> " << y << " ----: " << x << " ";
    while (mpz_cmp_ui(x,1)>0) // Teiler aus FB vorhanden?
     {
//       cout << "+";
       mpz_gcd(x,x,y); // enthlt die Primzahlen der FB, die gerade-oft y teilen
       mpz_divexact(y,y,x); // diese auch abdividieren
       mpz_mul(Delta,Delta,x); // und unter das Quadrat ziehen
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // teilt weiter...
        else
          WithOddExponent.push_back(FB_Primteiler[i]); // teilt hier letztmalig
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();

       mpz_gcd(x,x,y); mpz_divexact(y,y,x); // ungerade-oft
       for (unsigned int i=0; i<FB_Primteiler.size(); ++i)
        if (mpz_divisible_ui_p(x,PrimeNumbers[FB_Primteiler[i]]))
          TempTeiler.push_back(FB_Primteiler[i]); // teilt weiter...
       FB_Primteiler.swap(TempTeiler); TempTeiler.clear();
     }
    mpz_mod(Delta,Delta,n);
    std::sort(WithOddExponent.begin(),WithOddExponent.end());
    for (unsigned int i=0; i<WithOddExponent.size(); ++i)
     {
       Relation_sparse->append(WithOddExponent[i]);
//       cout << " " << WithOddExponent[i];
     }
//    cout << endl;
//    cout << "++++++> " << y << " ----: " << x << endl;

  }


  // reduce factors from dynamic factorbase (and maybe pushed squares...)
  // Faktoren zusammenmultiplizieren und dann in einem Rutsch abdividieren
  if (HitCount)
   {
     mpz_set_ui(x,Hits[0].Faktor); Hits[0].Exponent = 1;
     for (int i=1; i<HitCount; i++) { Hits[i].Exponent = 1; mpz_mul_ui(x,x,Hits[i].Faktor); }
#if defined(DEBUG)
     if (!mpz_divisible_p(y,x)) { MARK; cerr << "Wrong divisor in Hits!" << endl; exit(1); }
#endif
     mpz_divexact(y,y,x);
     mpz_gcd(x,x,y);
     if (mpz_cmp_ui(x,1)!=0) // are there higher exponents?
      {
        // find exponents of factors in dynamic factorbase (and maybe pushed squares...)
        for (int i=0; i<HitCount; ++i)
         {
           while (mpz_divisible_ui_p(y,Hits[i].Faktor))
     	    {
     	      mpz_divexact_ui(y,y,Hits[i].Faktor);
     	      Hits[i].Exponent++; // Exponent fr Faktor inkrementieren
     	    }
#ifdef DEBUG
           if (Hits[i].Exponent==2)
            {
              MARK;
              cout << "Square in Hits[]: " << Hits[i].Faktor;
              if (is_dynamic_factor(Hits[i].Faktor))
               cout << " (dynamic factor)" << endl;
              else
               cout << " (pushed square)" << endl;
            }
#endif
           if (Hits[i].Exponent>1)
     	    {
     	      mpz_set_ui(x, Hits[i].Faktor);
     	      mpz_powm_ui(x,x,Hits[i].Exponent/2, n);
     	      mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
     	      Hits[i].Exponent%=2;
     	    }
         }
      }
   }

  // Restfactor in static factorbase...
  const bool existiert_grosser_Restfaktor = (mpz_cmp_ui(y,biggest_Prime_in_Factorbase)>0);

  if (existiert_grosser_Restfaktor && collecting_phase_finished)
    { relevant_factor=special_factor; return; } // vorzeitiges Abbrechen

  if (!existiert_grosser_Restfaktor)
    {
      if (mpz_cmp_ui(y,1)!=0) 
        {
          cout << "CRelation::constructor:" << endl;
          cout << "Fehler in Relation / BUG im Programm:" << endl;
          cout << y << " should be 1!" << endl;
          exit(1);
        }
      relevant_factor=largest_factor_in_Relation();
      // mit der aufgenommenen Relation lt sich dieser Faktor eliminieren
    }

#if 0 || defined(SIEVING_MORE_LARGE_SQUARES)
  // dynamic factor sanity check
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
    {
      if (!is_dynamic_factor(Hits[i].Faktor))
	{
	  cerr << "Inconsistency in dynamic factorbase! " << Hits[i].Faktor << " not found as known dynamic factor!" << endl;
          // entweder tatschlich ein Fehler ODER
          // denkbarer (unwahrscheinlicher) Sonderfall:
          // ein gepushtes Primzahl-Quadrat auerhalb der dynamischen
          // Faktorbasis geht (sogar) mit ungerader Potenz ein...
          // ...dann liegt zwar kein Fehler vor, aber die Relation ist
          // trotzdem ungltig. Der Fall erscheint mir aber derart
          // unwahrscheinlich, dass ich dann trotzdem abbreche.
          // Hinzu kommt: Wenn ein Client einen solchen Faktor sieben sollte,
          // dann wird diese Bearbeitung hier gar nicht aufgerufen und
          // der Server wird im Fehlerfall die Relation nicht akzeptieren.
          // Dann gibt es zwar eine mgliche dynamische Relation weniger,
          // aber das sollte bei der Seltenheit dieses Falles nicht stren.
	  exit(1); // not necessary, because it should also be safe just to skip this relation
          relevant_factor=dynamic_factor; return; // skip the relation
	}
    }
#else
#ifdef DEBUG
 #warning "dynamic factor sanity check disabled! (just want to mention it)"
#endif
#endif

  if (existiert_grosser_Restfaktor)
    {
      const bool is_possible_dynamic = (mpz_cmp_ui(y, SingleLargePrime_Threshold)<=0);
      if ( !is_possible_dynamic && CmpqsFactor::DLP_rejected(y) )
        {
          // We should first filter out invalid big numbers, especially the
          // primes! It doesn't harm (except for performance) if some composite
          // numbers are filtered out as primes.
          // Important is that "DLP_reject(y)" is not too expensive!
	  relevant_factor=special_factor; // ignored factor (because we do not insert any relation)
	  return;
        }
        
      if (is_possible_dynamic && probab_prime(mpz_get_ui(y)) )
        {
          // es liegt ein neuer dynamic Factor vor
          //cout << "new dynamic-factor..." << endl;
          relevant_factor=dynamic_factor;
          TDynamicFactorRelation relation;
          relation.factor=mpz_get_ui(y);
          relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);

          // the found factor could be a "non-sievable" SingleLargePrime
          // We should detect this and take appropriate action!
          if (!relation.sieveable() && is_dynamic_factor(relation.factor))
           {
             Hits[HitCount].Faktor=relation.factor; Hits[HitCount].Exponent=1; ++HitCount;
             relevant_factor=largest_factor_in_Relation();
             goto done_with_grosser_Restfaktor;
           }

#if !defined(IS_SERVER) && !defined(USE_FAKEHEAP)
          compute_Deltas_for(relation, SieveOffset+PhysicalSieveSize);
#endif
      
#ifndef IS_CLIENT
          relation.fpos = save(DynamicRelations_to_file,relation.factor, HitCount);
#else
          // an das Ausgabefile
          relation.fpos=save(communication_stream,relation.factor, HitCount);
#endif
          DynamicFactorRelations.insert(relation);
          statistical_data::directly_sieved_DynamicFactors++;
#ifndef IS_CLIENT
          SpecialRelations::split_by_primefactor(relation.factor);
#endif
          return;
        }
        
      // es liegt ein Specialfactor vor, der nicht prim ist
      {
        relevant_factor=special_factor;
        TSpecialFactorRelation relation;
        if (!relation.factor.DLP_get(y))
          {
            //cerr << "Hint: Special Factor " << y << " invalid." << endl;
            return; /* Treffer nicht bernehmen, da er ungltig ist. */
          }
#ifndef IS_CLIENT
        if (!SpecialRelations::insert(relation.factor,this,HitCount))
          {
            relevant_factor=largest_factor_in_Relation();
            // wenn Factor nicht vermerkt wurde, heit das, dass er
            // neutralisiert werden konnte. -> normale Relation liegt vor.

            // /FIXME: nachfolgendes sollte vielleicht besser in SpecialRelations::insert() getan werden?
            // kombiniere Faktoren der dynamic Factorbase
            multi_combine_init();
            for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
             {
               //cout << "kombiniere dynamischen Factor " << Hits[i].Faktor << endl;
               TDynamicFactorRelation FaRelSearch;
               FaRelSearch.factor=Hits[i].Faktor;
               TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
               mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
               multi_combine_main(DynamicRelations_from_file,(*p).fpos);
               //cout << "fertig kombiniert von " << *pos << endl;
             }
            multi_combine_exit();
            return;
          }
#else
        // an das Ausgabefile
        relation.fpos=save(communication_stream, relation.factor, HitCount);
#endif
	statistical_data::directly_sieved_SpecialFactors++;
        return;
      }
    } // end of "if (existiert_grosser_Restfaktor)"
 done_with_grosser_Restfaktor:

#ifndef IS_CLIENT
  // kombiniere Faktoren der dynamic Factorbase
  multi_combine_init();
  for (int i=0; i<HitCount; ++i) if(Hits[i].Exponent!=0)
   {
     //cout << "kombiniere dynamischen Factor " << Hits[i].Faktor << endl;
     TDynamicFactorRelation FaRelSearch;
     FaRelSearch.factor=Hits[i].Faktor;
     TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
     mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
     multi_combine_main(DynamicRelations_from_file,(*p).fpos);
     //cout << "fertig kombiniert von " << *pos << endl;
   }
  multi_combine_exit();
#else
  // Ausgabe an File
  save(communication_stream,1, HitCount);
#endif
      
  if (HitCount<6) ++statistical_data::DynamicFactorRating[HitCount];
  else ++statistical_data::DynamicFactorRating[6];
  // bei DynamicFactorRating[0]: Relation vollstndig durch statische FB entdeckt
  // bei DynamicFactorRating[0<k<6]: Relation durch Eliminierung von (genau) k Faktoren aus dyn. FB entdeckt
  // bei DynamicFactorRating[6]: Relation durch Eliminierung von mindestens 6 Faktoren aus dyn. FB entdeckt
  
  //cout << "GL: " << *this << endl;
  //cout << "relevanter Faktor: " << relevant_factor << endl;
  //cout << "Initialisierung der Gleichungszeile abgeschlossen!" << endl;
}

#endif /* IS_SERVER not defined */



CmpqsFactor CRelation::combine(istream &in)
{
  //cout << "kombiniere Gleichungszeile vom File" << endl;
  CmpqsFactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  const ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x ist klassenweite Hilfsvariable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  mpz_set_ui(x,1); // zur Beschleunigung Faktoren in x aufsammeln, bis sie grer als n werden
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test(index))
	    if (index>0)
	      {
		mpz_mul_ui(x,x,PrimeNumbers[index]);
		if (mpz_cmp(x,n)>=0)
		  {
		    mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		    mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		  }
	      } //else mpz_neg(Delta,Delta);
	  Relation_dense->invert(index);
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto fertig;
	    }
	  while (index==R[p1])
	    {
	      if (index>0)
		{
		  mpz_mul_ui(x,x,PrimeNumbers[index]);
		  if (mpz_cmp(x,n)>=0)
		    {
		      mpz_mod(x,x,n); mpz_mul(Delta,Delta,x);
		      mpz_mod(Delta,Delta,n); mpz_set_ui(x,1);
		    }
		} //else mpz_neg(Delta,Delta);
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==R.size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
      //optisize(); // Speicherbedarf optimieren (sparse<->dense, etc.)
    }
  
  // x beinhaltet (mglicherweise) noch einen Restwert, der zu Delta multipliziert werden mu:
  mpz_mul(Delta,Delta,x); mpz_mod(Delta,Delta,n);
 
  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */
      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // Breitensuche, um dauerndes seeken zu vermeiden
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unknown correction value (Korrekturfaktor) " << FaRelSearch.factor << endl;
	      exit(1); // sicherheitshalber fr die partiell korrekte Variante...
	      ret=0; break; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (combine(DynamicRelations_from_file,(*p).fpos).IsTypeOf(CmpqsFactor::empty)) { ret=0; break; } // auf Hintergrundspeicher kombinieren und bei Fehler ret=0 setzen
	  //cout << "finished combining " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevant_factor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

CmpqsFactor CRelation::multi_combine_main(istream &in)
{
  //cout << "Multi-Kombiniere Gleichungszeilen vom File..." << endl;
  ++multi_combine_Counter; // Anzahl Aufrufe zhlen

  CmpqsFactor ret;
  string s;
  int index;
 
  // Dateiformatflags retten
  const ios::fmtflags oldFlags = in.flags(); // alte Formatflags retten
  in.unsetf(ios::hex); in.setf(ios::dec); // Dezimalausgabe sicherstellen (beachte: Rekursiver Aufruf in Lazy-Variante wegen Korrekturfaktoren mglich!)
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
 
  in >> s; // Startzeichen
  if (s != "G")
    {
      MARK;
      cerr << "error: wrong position while reading relation" << endl;
      exit(1);
    }
  in >> ret; // Faktor einlesen (dynamic-factor oder special-factor der Relation oder 1)
  in >> s; // delta
  mpz_set_str(x, s.c_str(), mpzbase_f); // x ist klassenweite Hilfsvariable
  mpz_mul(Delta, Delta, x);
  mpz_mod(Delta, Delta, n);

  // ab jetzt wird hexadezimal gearbeitet:
  in.unsetf(ios::dec); in.setf(ios::hex); // Hexadezimalausgabe
  in.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /*
    Symmetrische Differenz der Factors in Relation bilden.
    Gemeinsamen Anteil in einfacher(!!) Potenz in Relationswurzel aufnehmen
  */
  int distance; // Hinweis: statt der Werte werden deren Abstnde eingelesen...

  if (Relation_dense) /* Relation is dense */
    { /* symmetrische Differenz in dense-Darstellung bilden */
      distance=GetValue(in); index=distance;
      while (distance >= 0)
	{
	  if (Relation_dense->test_and_invert(index)) ++ExponentArray[index];
	  distance=GetValue(in); index+=distance;
	}
    }
  else
    { /* symmetrische Differenz in sparse-Darstellung bilden */
      int p1 = 0;
      static CTinyFBsizetypeVector neueMenge(StaticFactorbase::Size());
      neueMenge.reset(); // "leeren" ohne Speicherfreigabe
      const CTinyFBsizetypeVector& R = *Relation_sparse;
      
      distance=GetValue(in); index=distance; 
      while (distance>=0 && p1<R.size())
	{
	  while (index<R[p1])
	    {
	      neueMenge.append(index); distance=GetValue(in); index+=distance;
	      if (distance<0) goto fertig;
	    }
	  while (R[p1]<index)
	    {
	      neueMenge.append(R[p1]); ++p1;
	      if (p1==R.size()) goto fertig;
	    }
	  while (index==R[p1])
	    {
	      ++ExponentArray[index];
	      distance=GetValue(in); index+=distance; ++p1;
	      if (distance<0) goto fertig;
	      if (p1==R.size()) goto fertig;      
	    }
	}
    fertig:
      while (distance>=0) { neueMenge.append(index); distance=GetValue(in); index+=distance; }
      while (p1!=R.size()) { neueMenge.append(R[p1]); ++p1; }
      Relation_sparse->copy_from(neueMenge); // neue Menge bernehmen
    }

  if (distance == -2) // Korrekturfaktoren
    {
      /* Korrekturfaktoren sind dynamische Faktoren, die zum Zeitpunkt der Generierung
	 der gelesenen Relation nicht aufgelst wurden. Sie mssen daher jetzt verknpft werden. */

      TDynamicFactorRelation FaRelSearch;
      stack<int> DeferredFactors; // Breitensuche, um dauerndes seeken zu vermeiden
      while ( (FaRelSearch.factor=GetValue(in))>0 ) DeferredFactors.push(FaRelSearch.factor);
       
      while (!DeferredFactors.empty())
	{
	  FaRelSearch.factor=DeferredFactors.top(); DeferredFactors.pop();
	  //cout << "Korrekturfaktor: " << FaRelSearch.factor << " fr " << setprecision(20) << ret << endl;
	  TDynamicFactorRelations::iterator p = DynamicFactorRelations.find(FaRelSearch);
	  if (p==DynamicFactorRelations.end())
	    {
	      cerr << "Unknown correction value (Korrekturfaktor) " << FaRelSearch.factor << endl;
              exit(1); // fr die partiell korrekte Variante...
	      ret=0; break; // Weiterarbeiten trotzdem ermglichen... 
	    }
	  mpz_mul_ui(Delta,Delta,(*p).factor); mpz_mod(Delta,Delta,n);
	  if (multi_combine_main(DynamicRelations_from_file,(*p).fpos).IsTypeOf(CmpqsFactor::empty)) { ret=0; break; } // auf Hintergrundspeicher kombinieren; im Fehlerfall ret=0 setzen
	  //cout << "finished combining " << FaRelSearch.factor << endl;
	}
    }
  
  in.flags(oldFlags); // alte Formatflags des Streams wiederherstellen
 
  relevant_factor=largest_factor_in_Relation();
  //cout << "Kombinierung beendet." << endl;
  return ret;
}

streampos CRelation::save(ostream &out, const CmpqsFactor factor, const short int HitCount /* = 0 */) const
{
  //cout << "Speichern einer Gleichungszeile" << endl;
#ifndef IS_CLIENT
  streampos pos;
  // out.seekp(0,ios::end);  // sollte jetzt schon am Ende sein
  pos = out.tellp(); // fileposition ans Ende setzen und auslesen
#endif
  char *str;

  str = new char [mpz_sizeinbase(Delta,mpzbase_f)+2]; mpz_get_str(str,mpzbase_f, Delta);
  out << "G " << setprecision(20) << factor << " " << str << " ";
  delete [] str;

  // ab hier wird hexadezimal gearbeitet:
  const ios::fmtflags oldFlags = out.flags(); // alte Formatflags retten
  out << hex; // Hexadezimalausgabe
  out.unsetf(ios::showbase); // ohne Basiskennzeichnung
  
  /* Bemerkung:
     Statt der Werte werden die Abstnde gespeichert, das spart Platz */
  if (Relation_dense)
    { 
      int prev = 0;
      for (int i=Relation_dense->first(); i>=0; i=Relation_dense->next(i))
	{
          PutValue(out,i-prev);
	  prev=i;
	} 
    }
  else
    {
      int prev = 0;
      for (int i=0; i<Relation_sparse->size(); ++i)
	{
 	  int h = (*Relation_sparse)[i];
          PutValue(out,h-prev);
	  prev=h;
	}
    }

  int Anzahl_echte_Faktoren = 0;
  for (int i=0; i<HitCount; i++) if (Hits[i].Exponent>0) Anzahl_echte_Faktoren++;
  if (Anzahl_echte_Faktoren>0)
    {
      PutValue(out,-2);
      //cout << "Korrekturfaktoren fr Relation mit Faktor " << setprecision(20) << factor << endl;
      for (int i=0; i<HitCount; i++)
	{ 
	  //cout << Hits[i].Faktor << "^" << Hits[i].Exponent << endl; 
	  if (Hits[i].Exponent>0) PutValue(out,Hits[i].Faktor);
	}
    }
  PutValue(out,-1); out << endl;
  out.flags(oldFlags); // alte Formatflags wiederherstellen
#ifdef IS_CLIENT
  return 0;
#else
  return pos;
#endif
}



#ifndef IS_CLIENT
const bool CRelation::ComputeQuadraticCongruence() const
{
  /* Die ermittelte Relation erfllt die Kongruenz (A=1) mod n.
     Wegen A-1=(A+1)*(A-1)=0 mod n knnen wir daher nun durch ggT(A+1,n)
     einen Teiler von n ermitteln und haben gute Chancen, dass es sich dabei
     auch um einen echten (also nichttrivialen) Teiler handelt... */

  StatusReport(true);
  cout << endl << "Found a suitable relation..." << endl;

  if (!empty())
   {
     MARK;
     cerr << "Error! Relation contains unresolved members!" << endl;
   }

  cout << "[" << Delta << "] = 1 (mod N)" << endl;
   
  mpz_powm_ui(x,Delta,2,n);
  cout << "-> " << x << " = 1 (mod N)" << endl;
  
  if (mpz_cmp_ui(x,1)!=0) 
    {
      cout << "Error! Squares are NOT congruent!" << endl;
      exit(1);
    }
  
  mpz_add_ui(x,Delta,1);
  mpz_gcd(x,x,n); mpz_divexact(y,n,x); /* beachte: n und nicht kN
				          ist die eigentlich zu
				          faktorisierende Zahl! */
  
  if ((mpz_cmp_ui(x,1)==0) || (mpz_cmp_ui(y,1)==0))
    {
      cout << "factorization was trivial..." << endl << endl;
    }
  else
    {
      if (mpz_cmp(x,y)>0) mpz_swap(x,y); // we want x<y

      cout << endl << "factor found!" << endl;
      cout << x << " * ";
      cout << y << endl;
      cout << "= ";
      cout << n << endl;
      
      if (mpz_probab_prime_p(x,probab_prime_checks))
	{
	  mpz_divexact(n,n,x);
	  cout << x << " is a factor." << endl;
          std::ostringstream comment;
          comment << " [dmpqs]";
	  Factorization_to_file << MAL(x,1,comment) << flush;
	}
      else
      	if (Potenztest(x)) mpz_divexact(n,n,x); // ggf. Potenzen streichen...

      if (mpz_probab_prime_p(y,probab_prime_checks))
	{
	  mpz_divexact(n,n,y);
	  cout << y << " is a factor." << endl;
          std::ostringstream comment;
          comment << " [dmpqs]";
	  Factorization_to_file << MAL(y,1,comment) << flush;
	}
      else
	if (Potenztest(y)) mpz_divexact(n,n,y); // ggf. Potenzen streichen...
      

      if (mpz_cmp_ui(n,1)==0) // factorization complete?
	{ 
	  cout << "factorization successfully completed." << endl;
          Factorization_to_file << endl;
          return true; 
	}
    }
  cout << "continue factorization..." << endl << endl;
  return false;
}
#endif


// --------- Vorfaktorbestimmung -------------
// Um verschiedene Vorfaktorbestimmungsmethoden (mit jeweiliger Neu-Compilierung)
// testen zu knnen, ist die Prozedur " void determine_best_kN() "
// auf das File "Vorfaktorbestimmung.cc" ausgelagert.
#include "Vorfaktorbestimmung.cc"
// -------------------------------------------


#ifndef IS_CLIENT
 #include "parse_term.cc"
 // this is a predicate stating "false"
 const bool without_multi_combine_init = false; // true ->multi-init, false ->no multi_init on StaticRelations::insert
#endif

#include "SpecialRelations.cc"


#ifndef IS_SERVER
 // this part cannot be compiled separately and must therefore be included
 #include "Sieving-inc.cc"
#endif


#ifdef IS_STANDALONE
void process_ecm()
{
  streampos fpos = Recovery_from_file.tellg();
  int Anzahl_Kurven = 0;
  string s;

  srand(time(NULL)); // Random-Seed setzen (Zeitfunktion drfte hierzu zufllig genug sein)
  
  unsigned int dezimalstellen = mpz_sizeinbase(n,10);
  tune_parameters(dezimalstellen); // Parameter abhngig von der Gre der Zahl!

  Recovery_from_file >> Anzahl_Kurven;
  Recovery_from_file.ignore(1,'\n'); // endl-berlesen
  while (Anzahl_Kurven<elcu_Kurven)
  {

    Anzahl_Kurven++;
    cout << "elliptic curve #" << Anzahl_Kurven << "/" << elcu_Kurven << endl;

    int ecm_sigma = rand(); // zufllige Kurve whlen
    elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
    elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); /* Faktorisierung mit elliptischer Kurve versuchen */

    // Anzahl bearbeiteter Kurven ins Recovery-File schreiben...
    Recovery_to_file.seekp(fpos); 
    Recovery_to_file << Anzahl_Kurven << endl << flush;

    // Teiler gefunden?
    if (dezimalstellen==mpz_sizeinbase(n,10)) continue; // kein neuer Faktor
    
    // Stelligkeit von n hat sich gendert -> neuer Faktor

    if (mpz_probab_prime_p(n,probab_prime_checks))
      {
        Factorization_to_file << MAL(n);
        cout << "remaining factor: " << n << endl;
        cout << "remaining factor is most probably prime!" << endl;
        mpz_set_ui(n,1); // n ist faktorisiert
      }

    if ( (mpz_cmp_ui(n,1)==0) || Potenztest(n) ) // Faktorisierung vollstndig?
      { 
        cout << "factorization successfully completed." << endl;
        Factorization_to_file << endl;
        ExitManager::StopFactorization(); // Programm erfolgreich beenden...
      }

    // fr weitere ecm-Runden Parameter neu tunen...
    dezimalstellen=mpz_sizeinbase(n,10);
    tune_parameters(dezimalstellen);

    // nun mu noch das recovery-File angepat werden, da sich
    // die zu faktorisierende Zahl verkleinert hat!
    Recovery_buffer.close(); // File schlieen
    Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc); // wieder ffnen (leeres File erzeugen!)
    Recovery_to_file << n << endl; // zu faktorisierende Zahl ablegen
    fpos = Recovery_from_file.tellg(); // Position merken
    Recovery_to_file << Anzahl_Kurven << endl << flush; // Anzahl bisher bearbeiteter Kurven vermerken
    ecm_curves_processed=Anzahl_Kurven;

  }
  cout << "ECM-Phase (standalone) completed." << endl;
}
#endif



#ifdef IS_SERVER
 #include "Cprocess_clients.cc"
 #include "XML_StatusServer.cc"
#endif /* IS_SERVER */


#if defined(IS_CLIENT) && defined(USE_NETWORK)
 #include "Client_IO.cc"
#endif


#ifdef USE_NETWORK
void kill_all_other_threads()
{
  //cout << "Killing explicitely all other threads..." << endl;
  //pthread_kill_other_threads_np();
}
#endif

void cleanup_files()
{
  cout << "cleaning files..." << endl;
  StaticRelations::cleanup_files();
  SpecialRelations::cleanup_files();

  // close streams
  DynamicRelations_buffer.close();
  Recovery_buffer.close();

  // delete files
  remove(DynamicRelationsFile.c_str());
  remove(RecoveryFile.c_str());
}

void cleanup_memory()
{
  cout << "cleanup allocated memory" << endl;
  StaticRelations::cleanup_memory();
  cout << "cleanup mpz numbers" << endl;
  mpz_clear(r); mpz_clear(kN); mpz_clear(n);
  cout << "mpz numbers cleared." << endl;
}


#ifdef IS_SERVER
 void signalhandler(int signr)
 {
   cout << "Signal " << signr << " received. (ThreadId=" << pthread_self() << ")" << endl;
   exit(1);
 }
#endif


#ifdef USE_NCURSES
class Cncursed : private ForbidAssignment
{
private:
  static Cncursed* myself;
  static void suicide() { delete myself; }

  typedef streambuf* Pstreambuf;
  typedef Cwin* Pwin;
  Pstreambuf cout_titel_buf, cout_network_buf, cout_status_buf,
             cerr_buf, clog_buf, cout_buf;
  Pwin window1, window2, window3;
public:
 Cncursed()
 {
   if (myself)
    {
      MARK;
      cerr << "You may not call me twice, I'm unique!" << endl;
      exit(93);
    }
   myself=this; // I am myself, really!
   init_my_ncurses();
   window1 = new Cwin(8,0,0,0);
   window2 = new Cwin(8,0,8,0);
   window3 = new Cwin(0,0,16,0);
   cout_titel_buf=cout_titel.rdbuf(window1->rdbuf());
   cout_network_buf=cout_network.rdbuf(window1->rdbuf());
   cout_status_buf=cout_status.rdbuf(window2->rdbuf());
   cerr_buf=cerr.rdbuf(window2->rdbuf());
   clog_buf=clog.rdbuf(window2->rdbuf());
   cout_buf=cout.rdbuf(window3->rdbuf());
   atexit(suicide);
 }
 ~Cncursed()
 {
   cout_titel.rdbuf(cout_titel_buf);
   cout_network.rdbuf(cout_network_buf);
   cout_status.rdbuf(cout_status_buf);
   cerr.rdbuf(cerr_buf);
   clog.rdbuf(clog_buf);
   cout.rdbuf(cout_buf);
   delete window1; delete window2; delete window3;

   exit_my_ncurses();

   // print statistical information before exiting,
   // since screen should be cleared now!
   display_StatusLegend();
   StatusReport(true);
 }
};
Cncursed* Cncursed::myself = NULL;
#endif


int main(const int argc, const char* const argv[])
{

#ifdef USE_NETWORK
 atexit(kill_all_other_threads);
#endif

#ifdef USE_NCURSES
  new Cncursed(); // trigger activation of ncursed streams
#endif

  cout.setf(ios::fixed); // Dezimaldarstellung, nicht Scientific!

 {
   string name = "Faktorisierungsprogramm ";
   name+=VERSION;
   string author = "written 2004-08 by Thorsten Reinecke";
   string email = "(thre@thorstenreinecke.de)";
   const int L = 5+MAX(name.length(),MAX(author.length(),email.length()));
   const string sep(L,'-');
   int f;
   f=(L-name.length())>>1; name=string(f,' ')+name+string(f+1,' '); name.resize(L);
   f=(L-author.length())>>1; author=string(f,' ')+author+string(f+1,' '); author.resize(L);
   f=(L-email.length())>>1; email=string(f,' ')+email+string(f+1,' '); email.resize(L);
   cout_titel << " -" << sep << "-" << endl;
   cout_titel << "- " << name << " -" << endl;
   cout_titel << "- " << author << " -" << endl;
   cout_titel << "- " << email << " -" << endl;
   cout_titel << " -" << sep << "-" << endl;
 }
#ifdef __GNUG__
  {
    const string s=__VERSION__;
    cout << "(" << __DATE__ << ", " << __TIME__ << " using ";
    if (s[0]<='9') cout << "gcc ";
    cout << __VERSION__ << ")" << endl;
  }
#endif
  cout << endl;
  cout << "This program may be freely distributed under the" << endl;
  cout << "terms of the GNU-Public-Licence." << endl;
  cout << endl;
    
  if (argc>2)
    {
#ifdef IS_CLIENT
       cerr << "servername expected!" << endl;
#else
      cerr << "number for factorization expected!" << endl;
#endif
      exit(1);
    }
  
  const bool recover = (argc==1); // without argument: Recovery-Mode

  {
   // resolve ConfigFile
   if (getenv("QSIEVE_CFG"))
    {
      ConfigFile=getenv("QSIEVE_CFG");
      cout << "Setting ConfigFile to \"" << ConfigFile << "\"" << endl;
    }
   // and convert it to real/absolute path
#ifdef PATH_MAX
   char resolved_path[PATH_MAX];
#else
   char resolved_path[4096];
#endif /* PATH_MAX */
   if (realpath(ConfigFile.c_str(),resolved_path))
    ConfigFile=resolved_path;
   else
    {
      cerr << "Cannot find ConfigFile " << "\"" << ConfigFile << "\"" << endl;
      // server & standalone versions need a configuration file,
      // clients do not need them
#ifndef IS_CLIENT
      exit(1);
#endif
    }
  }

  mpz_init(n); /* Variable fr zu faktorisierende Zahl bereitstellen */
  mpz_init(kN);
  mpz_init(r);
  ExitManager::register_exithandler(cleanup_memory); // on sucessful exit free allocated data

#ifndef IS_CLIENT
  Read_ConfigFile();

#ifdef IS_SERVER
  // establish the network service at this point,
  // to allow clients to connect
  const connection_waiter my_connection_waiter(server_port);
#endif

  if (!recover) // Nicht im Recovery-Modus -> Argumente auswerten
    { 
      // bergebene Zahl holen (und ggf. ausrechnen)
      char* const neuer_str = strdup(argv[1]);
      char* str = neuer_str; 
      if (!parse_term::get_number(n, str))
	{
	  cout << "Wrong input at: '" << str << "'" << endl;
	  exit(1);
	}
      else
	if (str[0]!='\0')
	  {    
	    cout << "Syntax error in input term. (parenthesis?)" << endl;
	    exit(1);
	  }
	else
	  if (mpz_cmp_ui(n,0)<=0)
	    {
	      cout << "Input must be positive natural number!" << endl;
	      exit(1);
	    }
      free(neuer_str); // kein delete [] wegen stdup/malloc
      FoundFactors.regarding=argv[1];
    }

  if (recover)
    {
      // Recovery: continue a factorization, which was previously aborted
      // (try to) open the necessary file streams
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::in|ios::out);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }
      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::ate);

      // offenbar wird trotz ios::ate nicht (immer) ans Ende positioniert
      // deshalb wird die Testabfrage modifiziert:
      if (Recovery_from_file) Recovery_from_file.seekg(0,ios::end);
#ifdef STL_STREAM_workaround
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()==std::streampos(0)) || (Recovery_from_file.tellg()==std::streampos(-1)) )
// tellg()==0 indicates empty file -> we cannot recover
// tellg()==-1 indicates failure -> we cannot recover
// remark:
//  in gcc 3.4 (cvs-experimental-2003-10-17 we cannot compare with "<" !!)
//  do we really need a workaround to check this condition? 
#else
      if ( (!Recovery_from_file) || (Recovery_from_file.tellg()<1) )
#endif /* STL_STREAM_workaround */
	{
	  cerr << "Recovery not possible!" << endl;
	  exit(1);
	}
      
      Recovery_from_file.seekg(0,ios::beg);
      Recovery_to_file.seekp(0,ios::beg);
      Recovery_from_file >> n; // Zahl holen
      Recovery_from_file.ignore(1,'\n');

      cout << "Continue factorizaction for " << endl;
      cout << n << endl;
      // auerdem Recovery-Modus im Factorization-File vermerken, denn es
      // 1. knnte gelscht sein
      // 2. knnte der Recovery-Modus in seltenen Fllen Faktoren unterdrcken
      //    (wenn vorher das Schreiben unterbrochen wurde)
      // 3. knnte der Recovery-Modus zu doppelten Faktor-Ausgaben fhren
      //    (falls das quadratische Sieb einen zusammengesetzten Faktor
      //    gefunden hat und dann noch unterbrochen wird)
      // 4. knnte das Multiplikationssymbol inkorrekt hufig auftreten
      //    (wenn vorher kein Faktor gefunden wurde)
      Factorization_to_file << " [RECOVERY] "; // ohne flush, falls vor Finden weiterer Faktoren erneut abgebrochen wird... 
      FoundFactors.AutoLoad();
    }
  else
    {
      // Neustart: Starten einer neuen Faktorisierung
      // Streams ffnen
      StaticRelations::FileBuffer.open(StaticRelationsFile.c_str(),ios::out|ios::trunc);
      SpecialRelations::FileBuffer.open(SpecialRelationsFile.c_str(),ios::in|ios::out|ios::trunc);
      if (!DynamicRelations_buffer.open(DynamicRelationsFile.c_str(),ios::in|ios::out|ios::trunc))
       {
         cerr << "Cannot access " << DynamicRelationsFile << endl;
         exit(1);
       }

      Recovery_buffer.open(RecoveryFile.c_str(),ios::in|ios::out|ios::trunc);
      
      cout_status << "Starting factorization for" << endl;
      cout_status << n << endl;
      
      Factorization_to_file << endl << "Factorization of " << argv[1] << ":" << endl;
      Factorization_to_file << n << " = " << endl;
      
      easy_factor(); // "einfache" Faktoren rausstreichen
      
      cout_status << "Starting factorization with ECM and MPQS for" << endl;
      cout_status << n << endl;
      Recovery_to_file << n << endl;
    }
  ExitManager::register_exithandler(cleanup_files); // on sucessful exit delete the files
#ifdef IS_SERVER
  signal(SIGINT, signalhandler);
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning

#ifdef NOTIFY_PARENT
  // notify parent that now the distributed phase starts
  // and the server will process client connections
  kill(getppid(),SIGUSR1); // send SIGUSR1 to the parent process
#endif

  CXML_StatusServer::install_XML_StatusServer();
  if (!SkipECM)
   {
     cout << "Waiting for net-clients (ECM)..." << endl;
     Cprocess_clients::process_data_stream_ecm(my_connection_waiter); // erstmal ECM verteilt starten...
   }
#endif /* IS_SERVER */

#ifdef IS_STANDALONE
  // Auch die Standalone-Version sollte elliptische Kurven benutzen.
  if (!SkipECM) process_ecm(); // erstmal ECM starten...
#endif /* IS_STANDALONE */

#if defined(USE_DFT) || (CONTINUATION_METHOD==2)
  // an dieser Stelle sind in allen Programmvarianten die
  // Faktorisierungsalgorithmen, die discrete fast fourier transform
  // verwenden knnten, abgeschlossen.
  // Das DFT-Objekt kann also wieder freigegeben werden...
  polynomial::clear_dft_tempmemory(); // Lschen des DFT-Objektes triggern...
#endif /* USE_DFT */

  // nun wird alles fr das Quadratische Sieb vorbereitet...

  determine_best_kN(); /* optimaler Wert, fr den gesiebt werden kann */

  /* Streitfall:
     Soll man nach Ermittlung des bestem Multipliers die Siebparameter
     noch tunen?
     Ja   -> denn die Zahl kN hat ja die entsprechende Gre...
     Nein -> denn fr die Wahl des Multipliers wurden diese Gren nicht
             bercksichtigt und entscheidend sollte die Gre der
             zu faktorisierenden Zahl sein. Nur so ist eine Vergleichbarkeit
             der Multiplier unter gleichen Voraussetzungen zu gewhrleisten.
    Wie auch immer:
     Eine "optimale" Ermittlung der Siebgre und der Faktorbasis kann nur
     mittelbar von der Gre der Zahl abhngen. Die Laufzeit der Siebphase
     hngt davon ab, wie viele Relationen zu ermitteln sind *und* wie
     schnell die Relationen ermittelt (=gesiebt) werden knnen!
     Letzterer Wert ist funktional von der Gre der Zahl und der Qualitt
     der Faktorbasis abhngig. Diese Qualitt ist zwar durch Vorgabe eines
     Multipliers (nach Annahme) maximiert worden; aber das heit noch nicht,
     dass dadurch die funktionale Abhngigkeit beseitigt wre.
     Mit anderen Worten: Wenn wir die Bestimmung der Faktorbasisgre und
     der Siebgre *nur* von der Gre der Zahl abhngig machen, so gehen
     wir idealisierend von einer einheitlichen Qualitt der Faktorbasis aus.
     Da wir diese einheitliche Qualitt der Faktorbasis nicht definiert haben,
     besitzen wir auch keine Entscheidungsgrundlage darber, ob fr unsere
     Zahl (n bzw. kN) eine unterdurchschnittliche, durchschnittliche oder
     berdurchschnittliche Faktorbasisqualitt besteht. Aus dieser Sichtweise
     erscheint der obige Streitfall nicht entscheidbar zu sein.
  */

#if 0   
  tune_parameters(mpz_sizeinbase(kN,10)); // Parametertuning fr kN
#else
  tune_parameters(mpz_sizeinbase(n,10)); // Parametertuning fr n
#endif
  /* Hinweis:
      Beim Tuning der Parameter ist in der verteilten Version auf Konsistenz
      zu achten, d.h. ein Client sollte keine Parameter tunen!
      Unterschiedliche cfg-Files auf verschiedenen Rechnern wrden sonst
      zu Problemen fhren!
  */

#else /* IS_CLIENT */
  if (recover)
    {
      cerr << "Missing parameter" << endl;
      exit(1);
    }
  communication_name = argv[1];
#ifdef USE_NETWORK
  CClientRelation_Delivery::init();

  { // zunchst elliptische Kurven abarbeiten (net-client):
    int ecm_sigma = -1, runde=0;
    while (true)
     {
       unix_io_stream communication_stream(communication_name, server_port);
 waitloop:
       communication_stream << "ECM?" << endl;
       communication_stream >> ecm_sigma;
       if (ecm_sigma==-3)
        {
          // -3 is special token for idle waiting loop
          string s;
          communication_stream >> s;
          cout << s << endl;
          goto waitloop;
        }

       if (ecm_sigma<0) break; // sigma<0 -> keine Kurven mehr!
       communication_stream >> elcu_Phase1 >> elcu_Phase2;
       communication_stream >> n; // Zahl holen
       cout << "elliptic curve " << ++runde << endl;
       elliptic_curves elliptic_curve; // Instanz der Kurve kreieren
       elliptic_curve.go(ecm_sigma,elcu_Phase1,static_cast<elliptic_curves::NATNUM>(elcu_Phase2)); /* Faktorisierung mit elliptischer Kurve versuchen */
     }
  }
  
  {
    unix_io_stream communication_stream(communication_name, server_port);
    communication_stream << "Hunger!" << endl;
    communication_stream >> n;
    communication_stream >> StaticRelations::Size_StaticFactorbase;
    communication_stream >> Factor_Threshold;
    communication_stream >> LogicalSieveSize;
    cout << "n=" << n << endl;
    cout << "Size of static Faktorbase: " << StaticFactorbase::Size() << endl;
    cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
    cout << "Sieveinterval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
  }
  {
    string name;
    name = CClientRelation_Delivery::file_name+"A";
    communication_stream.open(name.c_str(), ios::out|ios::trunc);
  }
  atexit(CClientRelation_Delivery::cleanup_files); // Vorsicht: Bei Absturz sind dann noch nicht bertragene Relationen futsch!
#else /* no USE_NETWORK */
  // existiert das file?
  ifstream communication_stream2(communication_name.c_str());
  if (!communication_stream2)
    {
      cerr << "Unable to open " << communication_name << endl;
      exit (1);
    }
  communication_name+=".out";
  
  communication_stream2 >> n; 
  communication_stream2 >> StaticFactorbase::Size_StaticFactorbase;
  communication_stream2 >> Factor_Threshold;
  communication_stream2 >> LogicalSieveSize;
  communication_stream2.ignore(1,'\n');
  cout << "n=" << n << endl;
  cout << "Size of static Faktorbase: " << StaticFactorbase::Size() << endl;
  cout << "Factor-Threshold (Exponent): " << Factor_Threshold << endl;
  cout << "Sieveinterval per polynomial: [" << -LogicalSieveSize << "," << LogicalSieveSize << "]" << endl;
  communication_stream.open(communication_name.c_str(), ios::out|ios::trunc);
#endif /* USE_NETWORK */
  determine_best_kN(); /* optimaler Wert, fr den gesiebt werden kann */
#endif /* not IS_CLIENT */
  
  if (mpz_cmp_ui(r,PhysicalSieveSize)<0)
    {
      cerr << "Sieve size too big (you may want to reduce its size)!" << endl;
      exit(1);
    }

  // Fr Double-Large-Primes einen Schwellwert setzen,
  // dieser ist das Quadrat des maximalen Specialfactors...
  mpz_init(CmpqsFactor::DLP_Threshold);
  mpz_set_ui(CmpqsFactor::DLP_Threshold,SingleLargePrime_Threshold);
  mpz_mul(CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold,CmpqsFactor::DLP_Threshold);
  
  compute_StaticFactorbase();
  cout << "The first 20 members of the static factorbase are" << endl;
  for (int i=0; i<20; i++) cout << PrimeNumbers[i] << " "; cout << endl;
  cout << "Biggest prime in static factorbase: " << biggest_Prime_in_Factorbase << endl;
  cout << "#squares (of primes) in FB: " << FB_maxQuadrate << endl;
  cout << "#powers (of primes) in FB: " << NumberOf_more_PrimePowers << endl;

  // the server does not sieve anything, so initialize the sieve only, if
  // the program is not compiled in server mode
#ifndef IS_SERVER
  SieveControl::compute_SieveThreshold(); // Schwellwert zur Erkennung der Relationen fr Siebvorgang bestimmen
  for (int i=0; i<64; ++i) SieveArray_[i]=-1; // initialize array underflow trigger values
#endif


  Polynom.compute_first_polynomial(kN,LogicalSieveSize); // erstes Polynom ermitteln

#if defined(IS_CLIENT) && !defined(USE_NETWORK)
 // file-client: write kN to file
 communication_stream << kN << endl;
#endif
  
#ifndef IS_CLIENT
  if (recover)
    {
      // Static Relations einlesen
      cout << "Reading static relations from file..." << endl;
      StaticRelations::Load();
      cout << "Recover: " << StaticRelations::Count() << " static relations read." << endl;

      // Dynamic Relations einlesen
      cout << "Reading dynamic relations from file..." << endl;
      DynamicRelations_from_file.seekg(0,ios::beg);
      while (DynamicRelations_from_file.peek()=='G') 
	{
	  TDynamicFactorRelation relation;
	  relation.fpos = DynamicRelations_from_file.tellg();
	  string s; DynamicRelations_from_file >> s;
	  DynamicRelations_from_file >> relation.factor;
	  relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
	  DynamicRelations_from_file.ignore(1000000,'\n');
	  DynamicFactorRelations.insert(relation);
	}
      DynamicRelations_from_file.clear();
      DynamicRelations_to_file.seekp(0, ios::end); // sicherheitshalber...
      cout << "Recover: " << DynamicFactorRelations.size()
	   << " dynamic relations read." << endl;

      SpecialRelations::Load();
      SpecialRelations::CycleSearch();
      
      
#ifdef IS_SERVER
      // beim verteilten Sieben knnten den Clients einige potentielle dynamische Faktoren
      // durch die Lappen gegangen sein. Dies geschieht, falls ein Client einen Special-Factor
      // findet, der nicht mit dem vollen Satz der aktuellen (und daher nur dem Server bekannten)
      // dynamischen Faktoren gesiebt wurde.
      // Anstelle eines Tests zur Laufzeit werden diese im Normalfall eher seltenen Flle
      // nun hier beim Neu-Hochfahren des Servers bercksichtigt:
      // WICHTIG: erst mssen alle dynamischen Faktoren eingelesen sein! Ansonsten knnen
      //          doppelt auftretende dynamische Faktoren nicht zu einer Relation reduziert werden!
      // folgende Anweisungen wrden das bewerkstelligen, aber es lohnt sich kaum und dauert (zu) lange:
#if 1 /* 0: wir lassen es, 1: wir machen es (trotzdem) */
      cout << "Try to discover previously hidden special factors." << endl;
      for (TDynamicFactorRelations::iterator pos = DynamicFactorRelations.begin();
	   pos!=DynamicFactorRelations.end(); ++pos)
	SpecialRelations::split_by_primefactor((*pos).factor);
      cout << "discovered dynamic factors: " << statistical_data::Special_to_dynamic_Factor_hit << endl;
#endif
#endif
      
    }
#else // is client:
#ifndef USE_NETWORK 
  TDynamicFactorRelation relation;
  relation.fpos = 0; relation.factor=0;
  int diff; // wir lesen die Abstnde zwischen den Faktoren ein
  communication_stream2 >> diff; // filepos ueberlesen
  while (true)
    {
      communication_stream2 >> diff;
      relation.factor+=diff;
      // cout << "*" << relation.factor << "*" << flush;
      if (diff<=0) break;
      relation.SQRT_kN=SQRT_kN_mod_PrimeNumber(relation.factor);
      DynamicFactorRelations.insert(relation);
    }
  //cout << endl;
#endif
#endif  
  
  
  
#ifndef IS_CLIENT
  if (recover)
    {
      streampos fpos = Recovery_from_file.tellg();
      Polynom.load_if_available(Recovery_from_file); // aktuelles Polynom (zum Fortsetzen) ermitteln
      Recovery_to_file.seekp(fpos);
    }
#endif

display_StatusLegend();

#ifdef IS_SERVER
  cout << "Waiting for clients (MPQS)..." << endl;
  //cout << "I'm the main thread with id = " << pthread_self() << endl;
#if 0
  signal(SIGINT,  signalhandler);
  signal(SIGTSTP, signalhandler);
  signal(SIGQUIT, signalhandler);
  signal(SIGTTIN, signalhandler);
  signal(SIGTTOU, signalhandler);
  signal(SIGHUP,  signalhandler);
  signal(SIGCHLD, signalhandler);
#endif
  while (true)
   {    
     int* pdata = new int;
     *pdata = my_connection_waiter.get_new_client_socket_descriptor();
     pthread_attr_t detached_thread;
     pthread_t thread_process_client_data;

     pthread_attr_init(&detached_thread);
     pthread_attr_setdetachstate(&detached_thread, PTHREAD_CREATE_DETACHED);

     //cout << "creating thread..." << endl;
     const int retcode = pthread_create(&thread_process_client_data,&detached_thread,
                          Cprocess_clients::THREAD_process_data_stream, pdata);
     if (retcode != 0)
      {
        cerr << "pthread_create failed!" << endl;
	delete pdata;
        exit(1);
      }
      
     // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
     /* Durch das Flag "collecting_phase_finished" ist es mglich, das Aufsammeln weiterer dynamischer
        oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	Sammeln der eigentlichen Relationen macht, dass dynamische oder special-Faktoren hinderlich
	wirken.
	Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	des Programms wesentlich flexibler beeinflussen.
      */
   }

// Server
#else
// not Server


#ifdef IS_CLIENT
  mpz_t Obergrenze_D;
  mpz_init(Obergrenze_D);
    
#ifdef USE_NETWORK

 ClientMainLoop: // Hauptschleife: Polynomintervall holen und diese durchsieben

  CClientPolynomFetcher::fetch(Obergrenze_D);

#else // kein Netzwerk
  Polynom.load(communication_stream2);
  communication_stream2 >> Obergrenze_D;
  if (communication_stream.fail())
   {
     cerr << "No more data from file??" << endl;
     exit(1);
   }
#endif // if Netzwerk else keinNetzwerk
#endif // Client
  
  /* Quadratischen Reste mit Primzahlen aus der Primzahlbasis aussieben */
  do /* solange Relationen aussieben, bis mehr Relationen gefunden sind
	als Primzahlen in der Primzahlbasis sind (oder eine Faktorisierung
	gefunden ist). */
    {
#ifdef USE_NETWORK
      if (StaticRelations::Count() > 4*StaticFactorbase::Size())
       {
          cerr << "#static relations > 4*StaticFactorbase::Size(): this is weird." << endl;
          cerr << "I will sleep 30 seconds for transmission to end, and then I will abort." << endl;
          for (int i=0; i<15; ++i)
           {
             CClientRelation_Delivery::go();
             sleep(2);
           }
          exit(1);
       }
      CClientDynamicFactorFetcher::fetch();
#endif

#ifndef IS_CLIENT
      // fr Recovery:
      streampos fpos = Recovery_to_file.tellp();
      Polynom.save(Recovery_to_file); // aktuelles Polynom abspeichern
      Recovery_to_file.seekp(fpos);
#endif

      SieveOffset=-LogicalSieveSize;
      recompute_all_Deltas();
#ifdef USE_FAKEHEAP
      Sieve_Delta_Heap.sort();
#endif
      CSieveStaticFactors::init_LocalDeltas();
      while (SieveOffset<LogicalSieveSize)
	{
	  initialize_Sieve();
	  CSieveStaticFactors::do_sieving_StaticFactors(); // mit Primzahlen der statischen Faktorbasis
	  do_sieving_Squares(); // (grere) Quadrate von statischen Faktoren sieben
	  do_sieving_DynamicFactors(); // ermglicht weitere Relationen...
	  do_scanning_Sieve();
	  SieveOffset+=PhysicalSieveSize;
	}
      StatusReport(); // Ausgabe von Informationen

#if defined(IS_CLIENT) && defined(USE_NETWORK)
      CClientRelation_Delivery::go();
#endif  // Client && Network

      Polynom.compute_next_polynomial();
      
      // collecting_phase_finished= ( StaticRelations::Count() > 0.6*Size_StaticFactorbase );
      /* Durch das Flag "collecting_phase_finished" ist es mglich, das Aufsammeln weiterer dynamischer
	 oder special-Faktoren zu unterbinden. Dies knnte ratsam sein, wenn der Haupt- oder auch
	 der Hintergrundspeicher knapp wird oder das Programm schon solche Fortschritte im
	 Sammeln der eigentlichen Relationen macht, dass dynamische oder special-Faktoren hinderlich
	 wirken.
	 Da aber eine Speicherknappheit bei UNIX nicht sinnvoll ermittelt werden kann (andere Prozesse
	 knnen jederzeit Speicher anfordern/freigeben), empfiehlt es sich eher, das Programm
	 bei Speicherknappheit abzubrechen, die dynamic_relations-Datei zu krzen/zu lschen und
	 das Programm anschlieend neu zu starten. Auf diese Weise kann man die Speichernutzung
	 des Programms wesentlich flexibler beeinflussen.
      */
    }
#ifndef IS_CLIENT
  while (true);
#else
  while (Polynom < Obergrenze_D);
#ifdef USE_NETWORK
  goto ClientMainLoop; // Hauptschleife: neues Polynomintervall holen und fortsetzen
#endif
  mpz_clear (Obergrenze_D);
#endif
  
#endif /* not server */
  
  cout << endl << "Session ended successfully." << endl;
  return 0;
}
